# -*- coding: utf-8; -*-
"""add purchase_batch.po_total_calculated

Revision ID: 64d163d9de3a
Revises: abd43933c6c6
Create Date: 2019-06-07 16:44:15.007878

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '64d163d9de3a'
down_revision = 'abd43933c6c6'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch*
    op.add_column('purchase_batch', sa.Column('po_total_calculated', sa.Numeric(precision=8, scale=2), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('po_total_calculated', sa.Numeric(precision=7, scale=2), nullable=True))


def downgrade():

    # purchase_batch*
    op.drop_column('purchase_batch_row', 'po_total_calculated')
    op.drop_column('purchase_batch', 'po_total_calculated')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import os
import errno
import stat
import tempfile
from unittest import TestCase
from mock import patch, mock_open, call, DEFAULT

from rattail import daemon


class DaemonTests(TestCase):

    pidfile = '/tmp/test_rattail_daemon.pid'

    def test_init(self):
        d = daemon.Daemon(self.pidfile)
        self.assertEqual(d.pidfile, self.pidfile)
        self.assertEqual(d.stdin, os.devnull)
        self.assertEqual(d.stdout, os.devnull)
        self.assertEqual(d.stderr, os.devnull)

    def test_delpid(self):

        # normal delete
        with open(self.pidfile, 'w') as f:
            f.write('foo')
        self.assertTrue(os.path.exists(self.pidfile))
        d = daemon.Daemon(self.pidfile)
        d.delpid()
        self.assertFalse(os.path.exists(self.pidfile))

        # file doesn't exist (no error)
        d.delpid()

    @patch('sys.exit')
    @patch('sys.stderr')
    def test_start(self, stderr, exit_):
        d = daemon.Daemon(self.pidfile)
        with patch.object(d, 'daemonize') as daemonize:
            with patch.object(d, 'run') as run:

                # pidfile exists
                with open(self.pidfile, 'w') as f:
                    f.write('42\n')
                self.assertTrue(os.path.exists(self.pidfile))
                exit_.side_effect = RuntimeError
                self.assertRaises(RuntimeError, d.start)
                stderr.write.assert_called_once_with("pidfile /tmp/test_rattail_daemon.pid already exist. Daemon already running?\n")
                exit_.assert_called_once_with(1)
                self.assertFalse(daemonize.called)
                self.assertFalse(run.called)

                stderr.reset_mock()
                exit_.reset_mock()

                # no pidfile, no daemonize
                os.remove(self.pidfile)
                self.assertFalse(os.path.exists(self.pidfile))
                d.start(daemonize=False)
                self.assertFalse(stderr.write.called)
                self.assertFalse(exit_.called)
                self.assertFalse(d.daemonize.called)
                d.run.assert_called_once_with()

                d.run.reset_mock()

                # no pidfile, with daemonize
                self.assertFalse(os.path.exists(self.pidfile))
                d.start(daemonize=True)
                self.assertFalse(stderr.write.called)
                self.assertFalse(exit_.called)
                d.daemonize.assert_called_once_with()
                d.run.assert_called_once_with()


@patch('os.chdir')
@patch('os.chmod')
@patch('os.dup2')
@patch('os.fork')
@patch('os.getpid')
@patch('os.setsid')
@patch('os.umask')
@patch('sys.exit')
@patch('sys.stderr')
@patch('sys.stdin')
@patch('sys.stdout')
@patch('atexit.register')
class DaemonDaemonizeTests(TestCase):

    pidfile = '/tmp/test_rattail_daemon.pid'

    def tearDown(self):
        if os.path.exists(self.pidfile):
            os.remove(self.pidfile)

    def fake_open(self, *args):
        # Return mock values for stdin etc.
        if args[0] in ('/tmp/stdin', '/tmp/stdout', '/tmp/stderr'):
            return DEFAULT
        # Real value is returned for PID file.
        return open(*args)

    def test_first_fork_failure(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):
        fork.side_effect = OSError(errno.EIO, "I/O error")
        exit_.side_effect = RuntimeError
        d = daemon.Daemon(self.pidfile)
        self.assertRaises(RuntimeError, d.daemonize)
        stderr.write.assert_called_once_with("fork #1 failed: 5 (I/O error)\n")
        exit_.assert_called_once_with(1)

    def test_first_fork_success(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):
        fork.return_value = 42
        exit_.side_effect = RuntimeError
        d = daemon.Daemon(self.pidfile)
        self.assertRaises(RuntimeError, d.daemonize)
        self.assertFalse(stderr.write.called)
        exit_.assert_called_once_with(0)

    def test_second_fork_failure(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):
        fork.side_effect = [0, OSError(errno.EIO, "I/O error")]
        exit_.side_effect = RuntimeError
        d = daemon.Daemon(self.pidfile)
        self.assertRaises(RuntimeError, d.daemonize)
        stderr.write.assert_called_once_with("fork #2 failed: 5 (I/O error)\n")
        exit_.assert_called_once_with(1)

    def test_second_fork_success_parent(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):
        fork.side_effect = [0, 42]
        exit_.side_effect = RuntimeError
        d = daemon.Daemon(self.pidfile)
        self.assertRaises(RuntimeError, d.daemonize)
        self.assertFalse(stderr.write.called)
        exit_.assert_called_once_with(0)

    def test_second_fork_success_child(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):

        with patch('rattail.daemon.open', new=mock_open(), create=True) as open_:

            fork.side_effect = [0, 0]
            open_.side_effect = self.fake_open
            getpid.return_value = 42

            d = daemon.Daemon(self.pidfile,
                              stdin='/tmp/stdin',
                              stdout='/tmp/stdout',
                              stderr='/tmp/stderr')
            d.daemonize()

            stdout.flush.assert_called_once_with()
            stderr.flush.assert_called_once_with()

            self.assertEqual(open_.call_args_list, [
                    call('/tmp/stdin', 'r'),
                    call('/tmp/stdout', 'a+'),
                    call('/tmp/stderr', 'a+', 0),
                    call(self.pidfile, 'w+'),
                    ])

            stdin.fileno.assert_called_once_with()
            stdout.fileno.assert_called_once_with()
            stderr.fileno.assert_called_once_with()
            self.assertEqual(open_.return_value.fileno.call_count, 3)
            self.assertEqual(dup2.call_args_list, [
                    call(open_.return_value.fileno.return_value, stdin.fileno.return_value),
                    call(open_.return_value.fileno.return_value, stdout.fileno.return_value),
                    call(open_.return_value.fileno.return_value, stderr.fileno.return_value),
                    ])

            register.assert_called_once_with(d.delpid)
            getpid.assert_called_once_with()
            chmod.assert_called_once_with('/tmp/test_rattail_daemon.pid',
                                          stat.S_IRUSR | stat.S_IWUSR)

            with open(self.pidfile) as f:
                pid = f.read()
            self.assertEqual(pid, '42\n')

    def test_pidfile_error(self, register, stdout, stdin, stderr, exit_, umask, setsid, getpid, fork, dup2, chmod, chdir):

        tempdir = tempfile.mkdtemp()
        os.rmdir(tempdir)
        self.pidfile = os.path.join(tempdir, 'test_rattail_daemon.pid')
        self.assertFalse(os.path.exists(tempdir))
        self.assertFalse(os.path.exists(self.pidfile))

        with patch('rattail.daemon.open', new=mock_open(), create=True) as open_:
            fork.side_effect = [0, 0]
            open_.side_effect = self.fake_open

            d = daemon.Daemon(self.pidfile,
                              stdin='/tmp/stdin',
                              stdout='/tmp/stdout',
                              stderr='/tmp/stderr')
            self.assertRaises(IOError, d.daemonize)
            open_.assert_called_with(self.pidfile, 'w+')
            self.assertFalse(chmod.called)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Employment Handler
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.util import load_object
from rattail.app import GenericHandler


class EmploymentHandler(GenericHandler):
    """
    Base class and default implementation for employment handlers.
    """

    def begin_employment(self, person, start_date, **kwargs):
        """
        Begin employment for the given person.
        """
        session = self.get_session(person)

        # make sure we have an employee record
        employee = self.ensure_employee(person)
        session.flush()

        # employee status is now *current*
        employee.status = self.enum.EMPLOYEE_STATUS_CURRENT

        # maybe assign/update ID
        employee_id = kwargs.get('employee_id')
        if employee_id and employee.id != employee_id:
            employee.id = employee_id

        # create new history record, with start date
        history = self.make_employee_history(employee, start_date)
        session.flush()

        return employee

    def end_employment(self, employee, end_date, **kwargs):
        """
        End employment for the given employee.
        """
        session = self.get_session(employee)

        # employee status is now *former*
        employee.status = self.enum.EMPLOYEE_STATUS_FORMER

        # set end date for current history record, if present
        history = [h for h in employee.history if not h.end_date]
        if history:
            history = sorted(history, key=lambda h: h.start_date, reverse=True)
            history[0].end_date = end_date

    def ensure_employee(self, person):
        """
        Returns the employee record associated with the given person, creating
        it first if necessary.
        """
        employee = self.get_employee(person)
        if employee:
            return employee

        session = self.get_session(person)
        employee = self.make_employee(person)
        session.flush()
        return employee

    def get_employee(self, person):
        """
        Returns the employee associated with the given person, if there is one.
        """
        return person.employee

    def make_employee(self, person):
        """
        Create and return a new employee record.
        """
        employee = self.model.Employee()
        employee.person = person
        return employee

    def make_employee_history(self, employee, start_date):
        """
        Create and return a new employee history record.
        """
        history = self.model.EmployeeHistory()
        history.start_date = start_date
        employee.history.append(history)
        return history

    def why_not_begin_employment(self, person):
        """
        Inspect the given person and if they should not be made a current
        employee for any reason, return that reason as text.  If it's okay for
        the person to be made an employee, returns ``None``.
        """
        employee = self.get_employee(person)
        if employee and employee.status == self.enum.EMPLOYEE_STATUS_CURRENT:
            return "This person is already an employee"

    def why_not_end_employment(self, person):
        """
        Inspect the given person and if their current employment should not be
        ended for any reason, return that reason as text.  If it's okay for the
        person to stop being an employee, returns ``None``.
        """
        employee = self.get_employee(person)
        if not employee or employee.status != self.enum.EMPLOYEE_STATUS_CURRENT:
            return {'error': "This person is not currently an employee"}

    def get_context_employee(self, employee):
        """
        Return a dict of context data for the given employee.
        """
        status_display = None
        if employee.status is not None:
            status_display = self.enum.EMPLOYEE_STATUS.get(employee.status, "??")

        history = employee.sorted_history(reverse=True)
        if history:
            history = history[0]

        return {
            'uuid': employee.uuid,
            'id': employee.id,
            'status': employee.status,
            'status_display': status_display,
            'current': employee.status == self.enum.EMPLOYEE_STATUS_CURRENT,
            'start_date': six.text_type(history.start_date) if history else None,
            'end_date': six.text_type(history.end_date) if history and history.end_date else None,
        }


def get_employment_handler(config, **kwargs):
    """
    Create and return the configured :class:`EmploymentHandler` instance.
    """
    spec = config.get('rattail', 'employment.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = EmploymentHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Handler
"""

from __future__ import unicode_literals, absolute_import

import os
# import re
import tempfile

import six

from rattail.util import (load_object, load_entry_points,
                          OrderedDict, progress_loop, pretty_quantity)
from rattail.files import temp_path
from rattail.mail import send_email
from rattail.config import parse_list


class AppHandler(object):
    """
    Base class and default implementation for top-level Rattail app handler.

    aka. "the handler to handle all handlers"

    aka. "one handler to bind them all"
    """
    default_autocompleters = {
        'brands': 'rattail.autocomplete.brands:BrandAutocompleter',
        'customers': 'rattail.autocomplete.customers:CustomerAutocompleter',
        'customers.neworder': 'rattail.autocomplete.customers:CustomerNewOrderAutocompleter',
        'customers.phone': 'rattail.autocomplete.customers:CustomerPhoneAutocompleter',
        'employees': 'rattail.autocomplete.employees:EmployeeAutocompleter',
        'departments': 'rattail.autocomplete.departments:DepartmentAutocompleter',
        'people': 'rattail.autocomplete.people:PersonAutocompleter',
        'people.employees': 'rattail.autocomplete.people:PersonEmployeeAutocompleter',
        'people.neworder': 'rattail.autocomplete.people:PersonNewOrderAutocompleter',
        'products': 'rattail.autocomplete.products:ProductAutocompleter',
        'products.all': 'rattail.autocomplete.products:ProductAllAutocompleter',
        'products.neworder': 'rattail.autocomplete.products:ProductNewOrderAutocompleter',
        'vendors': 'rattail.autocomplete.vendors:VendorAutocompleter',
    }

    def __init__(self, config):
        self.config = config

        # app may not use the db layer, but if so we set the model
        try:
            self.model = config.get_model()
        except:
            pass

    def get_title(self, default=None):
        return self.config.app_title(default=default)

    def get_timezone(self, key='default'):
        from rattail.time import timezone
        return timezone(self.config, key)

    def localtime(self, *args, **kwargs):
        from rattail.time import localtime
        return localtime(self.config, *args, **kwargs)

    def make_utc(self, *args, **kwargs):
        from rattail.time import make_utc
        return make_utc(*args, **kwargs)

    def load_entry_points(self, group, **kwargs):
        return load_entry_points(group, **kwargs)

    def get_active_stores(self, session, **kwargs):
        """
        Returns the list of "active"
        :class:`~rattail.db.model.stores:Store` records.
        """
        import sqlalchemy as sa

        model = self.model
        return session.query(model.Store)\
                      .filter(sa.or_(
                          model.Store.archived == False,
                          model.Store.archived == None))\
                      .order_by(model.Store.id)\
                      .all()

    def get_autocompleter(self, key, **kwargs):
        """
        Returns a new :class:`~rattail.autocomplete.base.Autocompleter`
        instance corresponding to the given key, e.g. ``'products'``.

        The app handler has some hard-coded defaults for the built-in
        autocompleters (see ``default_autocompleters`` in the source
        code).  You can override any of these, and/or add your own
        with custom keys, via config, e.g.:

        .. code-block:: ini

           [rattail]
           autocomplete.products = poser.autocomplete.products:ProductAutocompleter
           autocomplete.otherthings = poser.autocomplete.things:OtherThingAutocompleter

        With the above you can then fetch your custom autocompleter with::

           autocompleter = app.get_autocompleter('otherthings')

        In any case if it can locate the class, it will create an
        instance of it and return that.

        :params key: Unique key for the type of autocompleter you
           need.  Often is a simple string, e.g. ``'customers'`` but
           sometimes there may be a "modifier" with it to get an
           autocompleter with more specific behavior.

           For instance ``'customers.phone'`` would effectively give
           you a customer autocompleter but which searched by phone
           number instead of customer name.

           Note that each key is still a simple string though, and that
           must be "unique" in the sense that only one autocompleter
           can be configured for each key.

        :returns: An :class:`~rattail.autocomplete.base.Autocompleter`
           instance if found, otherwise ``None``.
        """
        spec = self.config.get('rattail', 'autocomplete.{}'.format(key))
        if not spec:
            spec = self.default_autocompleters.get(key)
        if spec:
            return load_object(spec)(self.config)

        raise NotImplementedError("cannot locate autocompleter for key: {}".format(key))

    def get_auth_handler(self, **kwargs):
        if not hasattr(self, 'auth_handler'):
            spec = self.config.get('rattail', 'auth.handler',
                                   default='rattail.auth:AuthHandler')
            factory = load_object(spec)
            self.auth_handler = factory(self.config, **kwargs)
        return self.auth_handler

    def get_batch_handler(self, key, **kwargs):
        from rattail.batch import get_batch_handler
        return get_batch_handler(self.config, key, **kwargs)

    def get_board_handler(self, **kwargs):
        if not hasattr(self, 'board_handler'):
            from rattail.board import get_board_handler
            self.board_handler = get_board_handler(self.config, **kwargs)
        return self.board_handler

    def get_clientele_handler(self, **kwargs):
        if not hasattr(self, 'clientele_handler'):
            from rattail.clientele import get_clientele_handler
            self.clientele_handler = get_clientele_handler(self.config, **kwargs)
        return self.clientele_handler

    def get_employment_handler(self, **kwargs):
        if not hasattr(self, 'employment_handler'):
            from rattail.employment import get_employment_handler
            self.employment_handler = get_employment_handler(self.config, **kwargs)
        return self.employment_handler

    def get_feature_handler(self, **kwargs):
        if not hasattr(self, 'feature_handler'):
            from rattail.features import FeatureHandler
            self.feature_handler = FeatureHandler(self.config, **kwargs)
        return self.feature_handler

    def get_email_handler(self, **kwargs):
        if not hasattr(self, 'email_handler'):
            from rattail.mail import get_email_handler
            self.email_handler = get_email_handler(self.config, **kwargs)
        return self.email_handler

    # TODO: is it helpful to expose this? or more confusing?
    get_mail_handler = get_email_handler

    def get_all_import_handlers(self, ignore_errors=True, sort=False,
                                **kwargs):
        """
        Returns *all* Import/Export Handler classes which are known to
        exist, i.e.  all which are registered via ``setup.py`` for the
        various packages installed.

        This means it will include both "designated" handlers as well
        as non-designated.  See
        :meth:`get_designated_import_handlers()` if you only want the
        designated ones.

        Note that this will return the *Handler classes* and not
        *handler instances*.

        :param ignore_errors: Normally any errors which come up during
           the loading process are ignored.  Pass ``False`` here to
           force errors to raise, e.g. if you are not seeing the
           results you expect.

        :param sort: If you like the results can be sorted with a
           simple key based on "Source -> Target" labels.

        :returns: List of all registered Import/Export Handler classes.
        """
        # first load all "registered" Handler classes
        Handlers = self.load_entry_points('rattail.importing',
                                          ignore_errors=ignore_errors)

        # organize registered classes by spec
        specs = {}
        for Handler in six.itervalues(Handlers):
            spec = '{}:{}'.format(Handler.__module__, Handler.__name__)
            specs[spec] = Handler

        # many handlers may not be registered per se, but may be
        # designated via config.  so try to include those too
        for Handler in six.itervalues(Handlers):
            spec = self.get_designated_import_handler_spec(Handler.get_key())
            if spec and spec not in specs:
                specs[spec] = load_object(spec)

        # flatten back to simple list
        Handlers = list(specs.values())

        if sort:
            Handlers.sort(key=lambda h: (h.get_generic_host_title(),
                                         h.get_generic_local_title()))

        return Handlers

    def get_designated_import_handlers(self, with_alternates=False, **kwargs):
        """
        Returns all "designated" import/export handler instances.

        Each "handler type key" can have at most one Handler class
        which is "designated" in the config.  This method collects all
        registered handlers and then sorts out which one is
        designated, for each type key, ultimately returning only the
        designated ones.

        Note that this will return the *handler instances* and not
        *Handler classes*.

        If you have a type key and just need its designated handler,
        see :meth:`get_designated_import_handler()`.

        See also :meth:`get_all_import_handlers()` if you need all
        registered Handler classes.

        :param with_alternates: If you specify ``True`` here then each
           designated handler returned will have an extra attribute
           named ``alternate_handlers``, which will be a list of the
           other "available" (registered) handlers which match the
           designated handler's type key.

           This is probably most / only useful for the Configuration
           UI, to allow admin to change which is designated.

        :returns: List of all designated import/export handler instances.
        """
        grouped = OrderedDict()
        for Handler in self.get_all_import_handlers(**kwargs):
            key = Handler.get_key()
            grouped.setdefault(key, []).append(Handler)

        def find_designated(key, group):
            spec = self.get_designated_import_handler_spec(key)
            if spec:
                for Handler in group:
                    if Handler.get_spec() == spec:
                        return Handler

            if len(group) == 1:
                return group[0]

            raise RuntimeError("multiple ({}) handlers available but none designated for: {}".format(
                len(group), group[0].get_key()))

        designated = []
        for key, group in six.iteritems(grouped):
            Handler = find_designated(key, group)
            if Handler:
                # nb. we must instantiate here b/c otherwise if we
                # assign the `alternate_handlers` attr onto the class,
                # it can affect subclasses as well.  not so with
                # instances though
                handler = Handler(self.config)
                if with_alternates:
                    handler.alternate_handlers = [H for H in group
                                                  if H is not Handler]
                designated.append(handler)

        return designated

    def get_designated_import_handler(self, key, require=False, **kwargs):
        """
        Return the designated import/export handler instance, per the
        given handler type key.

        See also :meth:`get_designated_import_handlers()` if you want
        the full set of designated handlers.

        :param key: A "handler type key", e.g.
           ``'to_rattail.from_rattail.import'``.

        :param require: Specify ``True`` here if you want an error to
           be raised should no handler be found.

        :returns: The import/export handler instance corresponding to
           the given key.  If no handler can be found, then ``None``
           is returned, unless ``require`` param is true, in which
           case error is raised.
        """
        # first try to fetch the handler per designated spec
        spec = self.get_designated_import_handler_spec(key, require=require,
                                                       **kwargs)
        if spec:
            Handler = load_object(spec)
            return Handler(self.config)

        # nothing was designated, so leverage logic which already
        # sorts out which handler is "designated" for given key
        designated = self.get_designated_import_handlers(
            ignore_errors=not require)
        for handler in designated:
            if handler.get_key() == key:
                return handler

        if require:
            raise RuntimeError("Cannot locate designated handler for key: {}".format(
                key))

    def get_designated_import_handler_spec(self, key, require=False, **kwargs):
        spec = self.config.get('rattail.importing',
                               '{}.handler'.format(key))
        if spec:
            return spec

        legacy_setting = self.config.get('rattail.importing',
                                         '{}.legacy_handler_setting'.format(key))
        if legacy_setting:
            legacy_setting = parse_list(legacy_setting)
            if len(legacy_setting) == 2:
                section, option = legacy_setting
                spec = self.config.get(section, option)
                if spec:
                    return spec

        spec = self.config.get('rattail.importing',
                               '{}.default_handler'.format(key))
        if spec:
            return spec

        if require:
            raise RuntimeError("Cannot locate handler spec for key: {}".format(key))

    def get_membership_handler(self, **kwargs):
        """
        Returns a reference to the configured Membership Handler.

        See also :doc:`rattail-manual:base/handlers/other/membership`.
        """
        if not hasattr(self, 'membership_handler'):
            spec = self.config.get('rattail', 'membership.handler',
                                   default='rattail.membership:MembershipHandler')
            factory = load_object(spec)
            self.membership_handler = factory(self.config, **kwargs)
        return self.membership_handler

    def get_people_handler(self, **kwargs):
        """
        Returns a reference to the configured People Handler.

        See also :doc:`rattail-manual:base/handlers/other/people`.
        """
        if not hasattr(self, 'people_handler'):
            spec = self.config.get('rattail', 'people.handler',
                                   default='rattail.people:PeopleHandler')
            factory = load_object(spec)
            self.people_handler = factory(self.config, **kwargs)
        return self.people_handler

    def get_products_handler(self, **kwargs):
        if not hasattr(self, 'products_handler'):
            from rattail.products import get_products_handler
            self.products_handler = get_products_handler(self.config, **kwargs)
        return self.products_handler

    def get_report_handler(self, **kwargs):
        if not hasattr(self, 'report_handler'):
            from rattail.reporting import get_report_handler
            self.report_handler = get_report_handler(self.config, **kwargs)
        return self.report_handler

    def get_problem_report_handler(self, **kwargs):
        if not hasattr(self, 'problem_report_handler'):
            from rattail.problems import get_problem_report_handler
            self.problem_report_handler = get_problem_report_handler(
                self.config, **kwargs)
        return self.problem_report_handler

    def progress_loop(self, *args, **kwargs):
        return progress_loop(*args, **kwargs)

    def get_session(self, obj):
        """
        Returns the SQLAlchemy session with which the given object is
        associated.  Simple convenience wrapper around
        ``sqlalchemy.orm.object_session()``.
        """
        from sqlalchemy import orm

        return orm.object_session(obj)

    def make_session(self, **kwargs):
        """
        Creates and returns a new SQLAlchemy session for the Rattail DB.
        """
        from rattail.db import Session
        return Session(**kwargs)

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes
        :func:`rattail.db.cache.cache_model()` with the given model
        and keyword arguments.
        """
        from rattail.db import cache
        return cache.cache_model(session, model, **kwargs)

    def make_temp_dir(self, **kwargs):
        """
        Create a temporary directory.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkdtemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return tempfile.mkdtemp(**kwargs)

    def make_temp_file(self, **kwargs):
        """
        Reserve a temporary filename.  This is mostly a convenience wrapper
        around the built-in ``tempfile.mkstemp()``.
        """
        if 'dir' not in kwargs:
            workdir = self.config.workdir(require=False)
            if workdir:
                tmpdir = os.path.join(workdir, 'tmp')
                if not os.path.exists(tmpdir):
                    os.makedirs(tmpdir)
                kwargs['dir'] = tmpdir
        return temp_path(**kwargs)

    def normalize_phone_number(self, number):
        """
        Normalize the given phone number, to a "common" format that
        can be more easily worked with for sync logic etc.
        """
        from rattail.db.util import normalize_phone_number

        return normalize_phone_number(number)

    def phone_number_is_invalid(self, number):
        """
        This method should validate the given phone number string, and if the
        number is *not* considered valid, this method should return the reason
        as string.  If the number is valid, returns ``None``.
        """
        # strip non-numeric chars, and make sure we have 10 left
        normal = self.normalize_phone_number(number)
        if len(normal) != 10:
            return "Phone number must have 10 digits"

    def format_phone_number(self, number):
        """
        Returns a "properly formatted" string based on the given phone number.
        """
        from rattail.db.util import format_phone_number

        return format_phone_number(number)

    def render_currency(self, value, scale=2, **kwargs):
        """
        Must return a human-friendly display string for the given currency
        value, e.g. ``Decimal('4.20')`` becomes ``"$4.20"``.
        """
        if value is not None:
            if value < 0:
                fmt = "(${{:0,.{}f}})".format(scale)
                return fmt.format(0 - value)
            fmt = "${{:0,.{}f}}".format(scale)
            return fmt.format(value)

    def render_quantity(self, value, **kwargs):
        """
        Return a human-friendly display string for the given quantity
        value, e.g. ``1.000`` becomes ``"1"``.
        """
        return pretty_quantity(value, **kwargs)

    def render_cases_units(self, cases, units):
        """
        Render a human-friendly string showing the given number of
        cases and/or units.
        """
        if cases is not None:
            label = "case" if abs(cases) == 1 else "cases"
            cases = "{} {}".format(self.render_quantity(cases), label)

        if units is not None:
            label = "unit" if abs(units) == 1 else "units"
            units = "{} {}".format(self.render_quantity(units), label)

        if cases and units:
            return "{} + {}".format(cases, units)

        return cases or units

    def render_date(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``date``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d')

    def render_datetime(self, value, **kwargs):
        """
        Must return a human-friendly display string for the given ``datetime``
        object.
        """
        if value is not None:
            return value.strftime('%Y-%m-%d %I:%M:%S %p')

    def send_email(self, key, data={}, **kwargs):
        """
        Send an email message of the given type.

        See :func:`rattail.mail.send_email()` for more info.
        """
        send_email(self.config, key, data, **kwargs)


class GenericHandler(object):
    """
    Base class for misc. "generic" feature handlers.

    Most handlers which exist for sake of business logic, should inherit from
    this.
    """

    def __init__(self, config, **kwargs):
        self.config = config
        self.enum = self.config.get_enum()
        self.model = self.config.get_model()
        self.app = self.config.get_app()

    def progress_loop(self, *args, **kwargs):
        return self.app.progress_loop(*args, **kwargs)

    def get_session(self, obj):
        """
        Convenience wrapper around :meth:`AppHandler.get_session()`.
        """
        return self.app.get_session(obj)

    def make_session(self):
        """
        Convenience wrapper around :meth:`AppHandler.make_session()`.
        """
        return self.app.make_session()

    def cache_model(self, session, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.
        """
        return self.app.cache_model(session, model, **kwargs)


def make_app(config, **kwargs):
    """
    Create and return the configured :class:`AppHandler` instance.
    """
    spec = config.get('rattail', 'app.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = AppHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
"""
Alembic Environment Script
"""

from __future__ import unicode_literals, absolute_import

from alembic import context

from rattail.config import make_config
from rattail.db.config import get_default_engine


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
alembic_config = context.config

# Use same config file for Rattail, as we are for Alembic.
rattail_config = make_config(alembic_config.config_file_name, versioning=False)

# add your model's MetaData object here
# for 'autogenerate' support
from rattail.trainwreck.db.model import defaults as trainwreck
target_metadata = trainwreck.Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    engine = get_default_engine(rattail_config, section='trainwreck.db')
    context.configure(
        url=engine.url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = get_default_engine(rattail_config, section='trainwreck.db')
    connectable = engine
    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Trainwreck model importers
"""

from __future__ import unicode_literals, absolute_import

import datetime

from sqlalchemy import orm

from rattail import importing
from rattail.time import localtime, make_utc
from .util import ToOrFromTrainwreck


class ToTrainwreck(ToOrFromTrainwreck, importing.ToSQLAlchemy):
    """
    Base class for all Trainwreck model importers
    """
    key = 'uuid'


class TransactionImporter(ToTrainwreck):
    """
    Transaction data importer
    """
    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def cache_query(self):
        query = super(TransactionImporter, self).cache_query()
        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(self.model_class.system == system)
        return query.filter(self.model_class.end_time >= make_utc(self.start_time))\
                    .filter(self.model_class.end_time < make_utc(self.end_time))


class TransactionItemImporter(ToTrainwreck):
    """
    Transaction item data importer
    """

    @property
    def supported_fields(self):
        fields = super(TransactionItemImporter, self).supported_fields
        fields = list(fields)
        fields.extend([
            'transaction_system_id',
        ])
        return fields

    @property
    def importing_from_system(self):
        raise NotImplementedError("TODO: please define this for your subclass")

    def get_model_class(self):
        if hasattr(self, 'model_class') and self.model_class:
            return self.model_class
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.TransactionItem

    @property
    def transaction_class(self):
        trainwreck = self.config.get_trainwreck_model()
        return trainwreck.Transaction

    def setup(self):
        super(TransactionItemImporter, self).setup()

        if 'transaction_system_id' in self.fields:
            trainwreck = self.config.get_trainwreck_model()
            query = self.session.query(trainwreck.Transaction)
            try:
                system = self.importing_from_system
            except NotImplementedError:
                pass
            else:
                query = query.filter(trainwreck.Transaction.system == system)
            query = query.filter(trainwreck.Transaction.end_time >= make_utc(self.start_time))\
                         .filter(trainwreck.Transaction.end_time < make_utc(self.end_time))
            self.transactions_by_system_id = self.cache_model(trainwreck.Transaction,
                                                              query=query,
                                                              key='system_id')

    def normalize_local_object(self, item):
        data = super(TransactionItemImporter, self).normalize_local_object(item)
        if data:
            if 'transaction_system_id' in self.fields:
                data['transaction_system_id'] = item.transaction.system_id
            return data

    def cache_query(self):
        trainwreck = self.config.get_trainwreck_model()
        query = super(TransactionItemImporter, self).cache_query()
        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)
        return query.join(trainwreck.Transaction)\
                    .filter(trainwreck.Transaction.end_time >= make_utc(self.start_time))\
                    .filter(trainwreck.Transaction.end_time < make_utc(self.end_time))

    def get_transaction_by_system_id(self, system_id):
        if hasattr(self, 'transactions_by_system_id'):
            return self.transactions_by_system_id.get(system_id)

        trainwreck = self.config.get_trainwreck_model()
        query = self.session.query(trainwreck.Transaction)\
                            .filter(trainwreck.Transaction.system_id == system_id)
        try:
            system = self.importing_from_system
        except NotImplementedError:
            pass
        else:
            query = query.filter(trainwreck.Transaction.system == system)
        try:
            return query.one()
        except orm.exc.NoResultFound:
            pass

    def create_object(self, key, host_data):
        item = super(TransactionItemImporter, self).create_object(key, host_data)
        if item:

            # we may have to explicitly assign the transaction, if that uuid
            # wasn't part of our key
            # TODO: actually we do this if system_id was part of the key and do
            # not really look at uuid.  is there a better way to do or explain?
            # TODO: this also may fail outright if we get a bad system_id etc.
            if 'transaction_system_id' in self.key:
                system_id = host_data['transaction_system_id']
                txn = self.get_transaction_by_system_id(system_id)
                if not txn:
                    raise RuntimeError("transaction not found for system_id: {}".format(system_id))
                item.transaction = txn

            return item



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import shutil
import tempfile

from six.moves import queue
from unittest import TestCase

from rattail.config import make_config
from rattail.filemon import util
from rattail.filemon.config import Profile


class TestQueueExisting(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitor', u'foo')
        self.config.set('rattail.filemon', 'foo.dirs', self.tempdir)
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop')
        self.config.set(u'rattail.filemon', u'foo.action.noop.func', u'rattail.filemon.actions:noop')
        self.profile = Profile(self.config, u'foo')
        self.profile.queue = queue.Queue()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def write_file(self, fname, content):
        path = os.path.join(self.tempdir, fname)
        with open(path, 'wt') as f:
            f.write(content)
        return path

    def test_nothing_queued_if_no_files_exist(self):
        util.queue_existing(self.profile, self.tempdir)
        self.assertTrue(self.profile.queue.empty())

    def test_normal_files_are_queued_but_not_folders(self):
        self.write_file('file', '')
        os.makedirs(os.path.join(self.tempdir, 'folder'))
        util.queue_existing(self.profile, self.tempdir)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), os.path.join(self.tempdir, 'file'))
        self.assertTrue(self.profile.queue.empty())

    def test_if_profile_watches_locks_then_normal_files_are_queued_but_not_lock_files(self):
        self.profile.watch_locks = True
        self.write_file('file1.lock', '')
        self.write_file('file2', '')
        util.queue_existing(self.profile, self.tempdir)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), os.path.join(self.tempdir, 'file2'))
        self.assertTrue(self.profile.queue.empty())



# -*- coding: utf-8 -*-
"""add credit.vendor_item_code

Revision ID: 90a3338b187b
Revises: af4c4ec011fb
Create Date: 2018-03-01 11:23:19.029547

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '90a3338b187b'
down_revision = u'af4c4ec011fb'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_credit
    op.add_column('purchase_batch_credit', sa.Column('vendor_item_code', sa.String(length=20), nullable=True))

    # purchase_credit
    op.add_column('purchase_credit', sa.Column('vendor_item_code', sa.String(length=20), nullable=True))


def downgrade():

    # purchase_credit
    op.drop_column('purchase_credit', 'vendor_item_code')

    # purchase_batch_credit
    op.drop_column('purchase_batch_credit', 'vendor_item_code')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Importers
"""

from __future__ import unicode_literals, absolute_import

import logging

import six

from rattail.db import cache
from rattail.db.util import QuerySequence
from rattail.time import make_utc
from rattail.util import data_diffs, progress_loop, OrderedDict
from rattail.csvutil import UnicodeDictReader


log = logging.getLogger(__name__)


# TODO
class ImportLimitReached(Exception):
    pass


class Importer(object):
    """
    Base class for all data importers.

    .. attribute:: direction

       Should be a string, either ``'import'`` or ``'export'``.  This value is
       used to improve verbiage for logging and other output, for a better
       overall user experience.  It may also be used by importer logic, where
       the direction would otherwise be ambiguous.

       Note that the handler is responsible for assigning this value; the
       importer should not define it.  See also
       :attr:`rattail.importing.handlers.ImportHandler.direction`.

    :attr collect_changes_for_processing:

       If true (the default) then any changes occurring as a result of the
       import will be collected for processing by the handler, once the import
       has completed.  (I.e. it might send out a warning email with the
       changes.)  If the changes are not "important" per se, and they involve
       large data sets, then you may want to turn off this flag to avoid the
       overhead of collecting the changes.  In practice this is usually done if
       memory consumption is too great, as long as you don't actually need to
       track the changes.  Also note that the flag usually may be turned off 
       via command line kwarg (``--no-collect-changes``).
    """
    # Set this to the data model class which is targeted on the local side.
    model_class = None
    model_name = None

    key = None

    # The full list of field names supported by the importer, i.e. for the data
    # model to which the importer pertains.  By definition this list will be
    # restricted to what the local side can acommodate, but may be further
    # restricted by what the host side has to offer.
    supported_fields = []

    # The list of field names which may be considered "simple" and therefore
    # treated as such, i.e. with basic getattr/setattr calls.  Note that this
    # only applies to the local side, it has no effect on the host side.
    simple_fields = []

    allow_create = True
    allow_update = True
    allow_delete = True
    dry_run = False

    max_create = None
    max_update = None
    max_delete = None
    max_total = None
    batch_size = 200
    progress = None

    empty_local_data = False
    caches_local_data = False
    cached_local_data = None

    host_system_title = None
    local_system_title = None

    collect_changes_for_processing = True

    # TODO
    # Whether or not the registered "importer" batch handler is able to handle
    # batches for this importer (and/or, whether this importer is able to
    # provide what's needed for the same).
    batches_supported = False

    # TODO
    # If ``batches_supported`` is true, this should contain SQLAlchemy
    # ``Column`` instance overrides, keyed by fieldname.  Any field not
    # represented here will be given the default column type (string).
    field_coltypes = {}

    def __init__(self, config=None, key=None, direction='import', fields=None, exclude_fields=None, **kwargs):
        self.config = config
        self.app = config.get_app() if config else None
        self.enum = config.get_enum() if config else None
        self.model = config.get_model() if config else None
        self.model_class = kwargs.pop('model_class', self.get_model_class())
        if key is not None:
            self.key = key
        self.direction = direction
        self.fields = fields or self.supported_fields
        if exclude_fields:
            self.exclude_fields(*exclude_fields)
        if isinstance(self.key, six.string_types):
            self.key = (self.key,)
        if self.key:
            for field in self.key:
                if field not in self.fields:
                    raise ValueError("Key field '{}' must be included in effective fields "
                                     "for {}".format(field, self.__class__.__name__))
        self.model_name = kwargs.pop('model_name', self.model_name)
        if not self.model_name and self.model_class:
            self.model_name = self.model_class.__name__
        self._setup(**kwargs)

    def get_local_system_title(self):
        """
        Retrieve the system title for the local/target side.
        """
        if hasattr(self, 'handler'):
            return self.handler.local_title
        return self.local_system_title or "??"

    def include_fields(self, *args):
        """
        Add the given fields to the supported field list for the importer.  May
        be used at runtime to customize behavior.
        """
        for field in args:
            if field not in self.fields:
                self.fields.append(field)

    def exclude_fields(self, *args):
        """
        Remove the given fields from the supported field list for the importer.
        May be used at runtime to customize behavior.
        """
        for field in args:
            if field in self.fields:
                self.fields.remove(field)

    def get_model_class(self):
        return self.model_class

    def fields_active(self, fields):
        """
        Convenience method to check if any of the given fields are currently
        "active" for the importer.  Returns ``True`` or ``False``.
        """
        for field in fields:
            if field in self.fields:
                return True
        return False

    def _setup(self, **kwargs):
        self.create = kwargs.pop('create', self.allow_create) and self.allow_create
        self.update = kwargs.pop('update', self.allow_update) and self.allow_update
        self.delete = kwargs.pop('delete', False) and self.allow_delete
        for key, value in kwargs.items():
            setattr(self, key, value)

    def setup(self):
        """
        Perform any setup necessary, e.g. cache lookups for existing data.
        """

    def datasync_setup(self):
        """
        Perform any setup necessary, in the context of a datasync job.
        """

    def teardown(self):
        """
        Perform any cleanup after import, if necessary.
        """

    def progress_loop(self, func, items, factory=None, **kwargs):
        factory = factory or self.progress
        return progress_loop(func, items, factory, **kwargs)

    def unique_data(self, host_data):
        # Prune duplicate keys from host/source data.  This is for the sake of
        # sanity since duplicates typically lead to a ping-pong effect, where a
        # "clean" (change-less) import is impossible.
        unique = OrderedDict()
        for data in host_data:
            key = self.get_key(data)
            if key in unique:
                log.warning("duplicate records detected from {} for key: {}".format(
                    self.host_system_title, key))
            else:
                unique[key] = data
        return list(unique.values()), unique

    def import_data(self, host_data=None, now=None, **kwargs):
        """
        Import some data!  This is the core body of logic for that, regardless
        of where data is coming from or where it's headed.  Note that this
        method handles deletions as well as adds/updates.
        """
        self.now = now or make_utc(tzinfo=True)
        if kwargs:
            self._setup(**kwargs)
        self.setup()
        created = []
        updated = []
        deleted = []

        # Get complete set of normalized host data.
        if host_data is None:
            host_data = self.normalize_host_data()
        host_data, unique = self.unique_data(host_data)

        # Cache local data if appropriate.
        if self.caches_local_data:
            self.cached_local_data = self.cache_local_data(host_data)

        # Create and/or update data.
        if self.create or self.update:
            if self.collect_changes_for_processing:
                created, updated = self._import_create_update(host_data)
            else:
                self._import_create_update(host_data)

        # Delete data.
        if self.delete:
            changes = len(created) + len(updated)
            if self.max_total and changes >= self.max_total:
                log.warning("max of {} total changes already reached; skipping deletions".format(self.max_total))
            elif self.collect_changes_for_processing:
                deleted = self._import_delete(host_data, set(unique), changes=changes)
            else:
                self._import_delete(host_data, set(unique), changes=changes)

        self.teardown()
        return created, updated, deleted

    def _import_create_update(self, data, created=None, updated=None):
        """
        Import the given data; create and/or update records as needed.
        """
        if created is None:
            created = []
        if updated is None:
            updated = []
        count = len(data)
        if not count:
            return created, updated
        dummy = object()

        def import_(host_data, i):
            
            # Fetch local object, using key from host data.
            key = self.get_key(host_data)
            local_object = self.get_local_object(key)

            # If we have a local object, but its data differs from host, update it.
            if local_object and self.update:
                local_data = self.normalize_local_object(local_object)
                diffs = self.data_diffs(local_data, host_data)
                if diffs:
                    log.debug("fields '{}' differed for local data: {}, host data: {}".format(
                        ','.join(diffs), local_data, host_data))
                    local_object = self.update_object(local_object, host_data, local_data=local_data)
                    if self.collect_changes_for_processing:
                        updated.append((local_object, local_data, host_data))
                    else:
                        updated.append(dummy)
                    if self.max_update and len(updated) >= self.max_update:
                        log.warning("max of {} *updated* records has been reached; stopping now".format(self.max_update))
                        raise ImportLimitReached()
                    if self.max_total and (len(created) + len(updated)) >= self.max_total:
                        log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
                        raise ImportLimitReached()

            # If we did not yet have a local object, create it using host data.
            elif not local_object and self.create:
                local_object = self.create_object(key, host_data)
                if local_object:
                    log.debug("created new {} {}: {}".format(self.model_name, key, local_object))
                    if self.collect_changes_for_processing:
                        created.append((local_object, host_data))
                    else:
                        created.append(dummy)
                    if self.caches_local_data and self.cached_local_data is not None:
                        self.cached_local_data[key] = {'object': local_object, 'data': self.normalize_local_object(local_object)}
                    if self.max_create and len(created) >= self.max_create:
                        log.warning("max of {} *created* records has been reached; stopping now".format(self.max_create))
                        raise ImportLimitReached()
                    if self.max_total and (len(created) + len(updated)) >= self.max_total:
                        log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
                        raise ImportLimitReached()
                else:
                    log.debug("did NOT create new {} for key: {}".format(self.model_name, key))

            # flush changes every so often
            flush = False
            if not self.batch_size or self.batch_size == 1:
                # flush every time if no meaningful batch size
                flush = True
            elif created or updated:
                if (len(created) + len(updated)) % self.batch_size == 0:
                    flush = True
            if flush:
                self.flush_create_update()

        try:
            self.progress_loop(import_, data, message="{}ing {} data to {}".format(
                self.direction.capitalize(), self.model_name, self.get_local_system_title()))
        except ImportLimitReached:
            pass
        self.flush_create_update_final()
        return created, updated

    # def _populate_create_update(self, row_table, data):
    #     """
    #     Populate create and/or update records for the given batch row table,
    #     according to the given host data set.
    #     """
    #     created = []
    #     updated = []
    #     # count = len(data)
    #     # if not count:
    #     #     return created, updated

    #     def record(host_data, i):
            
    #         # fetch local object, using key from host data
    #         key = self.get_key(host_data)
    #         local_object = self.get_local_object(key)

    #         # if we have a local object, but its data differs from host, make an update record
    #         if local_object and self.update:
    #             local_data = self.normalize_local_object(local_object)
    #             diffs = self.data_diffs(local_data, host_data)
    #             if diffs:
    #                 log.debug("fields '{}' differed for local data: {}, host data: {}".format(
    #                     ','.join(diffs), local_data, host_data))
    #                 local_object = self.update_object(local_object, host_data, local_data)
    #                 updated.append((local_object, local_data, host_data))
    #                 if self.max_update and len(updated) >= self.max_update:
    #                     log.warning("max of {} *updated* records has been reached; stopping now".format(self.max_update))
    #                     raise ImportLimitReached()
    #                 if self.max_total and (len(created) + len(updated)) >= self.max_total:
    #                     log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
    #                     raise ImportLimitReached()

    #         # if we did not yet have a local object, make a create record
    #         elif not local_object and self.create:
    #             local_object = self.create_object(key, host_data)
    #             if local_object:
    #                 log.debug("created new {} {}: {}".format(self.model_name, key, local_object))
    #                 created.append((local_object, host_data))
    #                 if self.caches_local_data and self.cached_local_data is not None:
    #                     self.cached_local_data[key] = {'object': local_object, 'data': self.normalize_local_object(local_object)}
    #                 if self.max_create and len(created) >= self.max_create:
    #                     log.warning("max of {} *created* records has been reached; stopping now".format(self.max_create))
    #                     raise ImportLimitReached()
    #                 if self.max_total and (len(created) + len(updated)) >= self.max_total:
    #                     log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
    #                     raise ImportLimitReached()
    #             else:
    #                 log.debug("did NOT create new {} for key: {}".format(self.model_name, key))

    #         # flush changes every so often
    #         if not self.batch_size or (len(created) + len(updated)) % self.batch_size == 0:
    #             self.flush_create_update()

    #     try:
    #         self.progress_loop(record, data, message="Importing {} data".format(self.model_name))
    #     except ImportLimitReached:
    #         pass
    #     # self.flush_create_update_final()
    #     return created, updated

    def flush_create_update(self):
        """
        Perform any steps necessary to "flush" the create/update changes which
        have occurred thus far in the import.
        """

    def flush_create_update_final(self):
        """
        Perform any final steps to "flush" the created/updated data here.
        """
        self.flush_create_update()

    def _import_delete(self, host_data, host_keys, changes=0):
        """
        Import deletions for the given data set.
        """
        deleted = []
        deleting = self.get_deletion_keys() - host_keys
        count = len(deleting)
        log.debug("found {} instances to delete".format(count))
        dummy = object()

        def delete(key, i):
            cached = self.cached_local_data.pop(key)
            obj = cached['object']
            if self.delete_object(obj):
                if self.collect_changes_for_processing:
                    deleted.append((obj, cached['data']))
                else:
                    deleted.append(dummy)
                if self.max_delete and len(deleted) >= self.max_delete:
                    log.warning("max of {} *deleted* records has been reached; stopping now".format(self.max_delete))
                    raise ImportLimitReached()
                if self.max_total and (changes + len(deleted)) >= self.max_total:
                    log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
                    raise ImportLimitReached()

        try:
            self.progress_loop(delete, sorted(deleting),
                               message="Deleting {} data".format(self.model_name))
        except ImportLimitReached:
            pass

        return deleted

    def get_key(self, data):
        """
        Return the key value for the given data dict.
        """
        return tuple(data[k] for k in self.key)

    def get_host_objects(self):
        """
        Return the "raw" (as-is, not normalized) host objects which are to be
        imported.  This may return any sequence-like object, which has a
        ``len()`` value and responds to iteration etc.  The objects contained
        within it may be of any type, no assumptions are made there.  (That is
        the job of the :meth:`normalize_host_data()` method.)
        """
        return []

    def normalize_host_data(self, host_objects=None):
        """
        Return a normalized version of the full set of host data.  Note that
        this calls :meth:`get_host_objects()` to obtain the initial raw
        objects, and then normalizes each object.  The normalization process
        may filter out some records from the set, in which case the return
        value will be smaller than the original data set.
        """
        if host_objects is None:
            host_objects = self.get_host_objects()
        normalized = []

        def normalize(obj, i):
            data = self.normalize_host_object(obj)
            if data:
                normalized.append(data)

        self.progress_loop(normalize, host_objects,
                           message="Reading {} data from {}".format(
                               self.model_name, self.host_system_title))
        return normalized

    def normalize_host_object(self, obj):
        """
        Normalize a raw host object into a data dict, or return ``None`` if the
        object should be ignored for the importer's purposes.
        """
        return obj

    def get_local_objects(self, host_data=None):
        """
        Fetch all raw objects from the local system.
        """
        raise NotImplementedError

    def cache_local_data(self, host_data=None):
        """
        Cache all raw objects and normalized data from the local system.
        """
        objects = self.get_local_objects(host_data=host_data)
        cached = {}

        def cache(obj, i):
            data = self.normalize_local_object(obj)
            key = self.get_key(data)
            cached[key] = {'object': obj, 'data': data}

        self.progress_loop(cache, objects,
                           message="Reading {} data from {}".format(
                               self.model_name, self.get_local_system_title()))
        return cached

    def cache_local_message(self):
        """
        Must return a message to be used for progress when fetching "local" data.
        """
        return "Reading {} data from {}".format(self.model_name, self.get_local_system_title())

    def get_cache_key(self, obj, normal):
        """
        Get the primary cache key for a given object and normalized data.

        Note that this method's signature is designed for use with the
        :func:`rattail.db.cache.cache_model()` function, and as such the
        ``normal`` parameter is understood to be a dict with a ``'data'`` key,
        value for which is the normalized data dict for the raw object.
        """
        return tuple(normal['data'].get(k) for k in self.key)

    def normalize_cache_object(self, obj, data=None):
        """
        Normalizer for cached local data.  This returns a simple dict with
        ``'object'`` and ``'data'`` keys; values are the raw object and its
        normalized data dict, respectively.
        """
        if data is None:
            data = self.normalize_local_object(obj)
        return {'object': obj, 'data': data}

    def normalize_local_object(self, obj):
        """
        Normalize a local (raw) object into a data dict.
        """
        fields = [f for f in self.simple_fields
                  if f in self.fields]
        # note, we normally should have a proper object here, but in
        # some (e.g. dry-run) cases we may just have a dict; we should
        # do the right thing if so
        if isinstance(obj, dict):
            data = dict([(field, obj[field])
                         for field in fields])
        else:
            data = dict([(field, getattr(obj, field))
                         for field in fields])
        return data

    def get_local_object(self, key):
        """
        Must return the local object corresponding to the given key, or
        ``None``.  Default behavior here will be to check the cache if one is
        in effect, otherwise return the value from
        :meth:`get_single_local_object()`.
        """
        if not self.empty_local_data:
            if self.caches_local_data and self.cached_local_data is not None:
                data = self.cached_local_data.get(key)
                return data['object'] if data else None
            return self.get_single_local_object(key)

    def get_single_local_object(self, key):
        """
        Must return the local object corresponding to the given key, or None.
        This method should not consult the cache; that is handled within the
        :meth:`get_local_object()` method.
        """
        raise NotImplementedError

    def cache_model(self, model, **kwargs):
        """
        Convenience method which invokes :func:`rattail.db.cache.cache_model()`
        with the given model and keyword arguments.  It will provide the
        ``session`` and ``progress`` parameters by default, setting them to the
        importer's attributes of the same names.
        """
        session = kwargs.pop('session', None)
        if not session:
            session = self.session
        kwargs.setdefault('progress', self.progress)
        return cache.cache_model(session, model, **kwargs)

    def data_diffs(self, local_data, host_data):
        """
        Find all (relevant) fields which differ between the host and local data
        values for a given record.
        """
        return data_diffs(local_data, host_data, fields=self.fields)

    def make_object(self):
        """
        Make a new/empty local object from scratch.
        """
        return self.model_class()

    def new_object(self, key):
        """
        Return a new local object to correspond to the given key.  Note that
        this method should only populate the object's key, and leave the rest
        of the fields to :meth:`update_object()`.
        """
        obj = self.make_object()
        for i, k in enumerate(self.key):
            if hasattr(obj, k):
                setattr(obj, k, key[i])
        return obj

    def create_object(self, key, host_data):
        """
        Create and return a new local object for the given key, fully populated
        from the given host data.  This may return ``None`` if no object is
        created.
        """
        if not host_data.get('_deleted_', False):
            obj = self.new_object(key)
            if obj:
                return self.update_object(obj, host_data)

    def update_object(self, obj, host_data, local_data=None, all_fields=False):
        """
        Update the local data object with the given host data, and return the
        object.
        """
        for field in self.simple_fields:
            if field not in self.key and field in host_data and (all_fields or field in self.fields):
                if not local_data or field not in local_data or local_data[field] != host_data[field]:
                    setattr(obj, field, host_data[field])
        return obj

    def get_deletion_keys(self):
        """
        Return a set of keys from the *local* data set, which are eligible for
        deletion.  By default this will be all keys from the local cached data
        set, or an empty set if local data isn't cached.
        """
        if not self.caches_local_data:
            return set()

        if self.cached_local_data is None:
            return set()

        all_keys = set(self.cached_local_data)
        keys = set()

        def check(key, i):
            data = self.cached_local_data[key]['data']
            obj = self.cached_local_data[key]['object']
            if self.can_delete_object(obj, data):
                keys.add(key)

        self.progress_loop(check, all_keys,
                           message="Determining which objects can be deleted")
        return keys

    def can_delete_object(self, obj, data):
        """
        Should return a boolean indiciating whether or not the given
        object "can" be deleted.  Default is to return ``True`` in all
        cases.

        If you return ``False`` then the importer will not perform any
        delete action on the object.
        """
        return True

    def delete_object(self, obj):
        """
        Delete the given object from the local system (or not), and return a
        boolean indicating whether deletion was successful.  What exactly this
        entails may vary; default implementation does nothing at all.
        """
        return True

    def prioritize_2(self, data, field, field2=None):
        """
        Prioritize the data values for the pair of fields implied by the given
        fieldname.  I.e., if only one non-empty value is present, make sure
        it's in the first slot.
        """
        if not field2:
            field2 = '{}_2'.format(field)
        if field in data and field2 in data:
            if data[field2] and not data[field]:
                data[field], data[field2] = data[field2], None


class FromCSV(Importer):
    """
    Generic base class for importers whose data source is a CSV file.
    """

    def setup(self):
        if not hasattr(self, 'source_data_path'):
            if hasattr(self, 'args') and hasattr(self.args, 'source_csv'):
                self.source_data_path = self.args.source_csv

    def get_host_objects(self):
        if six.PY3:
            source_csv_file = open(self.source_data_path, 'rt', encoding='latin_1')
            reader = UnicodeDictReader(source_csv_file)
        else:
            source_csv_file = open(self.source_data_path, 'rb')
            reader = UnicodeDictReader(source_csv_file, 'latin_1')

        objects = list(reader)
        source_csv_file.close()
        return objects


class FromQuery(Importer):
    """
    Generic base class for importers whose raw external data source is a
    SQLAlchemy (or Django, or possibly other?) query.
    """

    def query(self):
        """
        Subclasses must override this, and return the primary query which will
        define the data set.
        """
        raise NotImplementedError

    def get_host_objects(self):
        """
        Returns (raw) query results as a sequence.
        """
        return QuerySequence(self.query())


class FromDjango(FromQuery):
    """
    Base class for importers using Django ORM on the host side.
    """
    django_dbkey = 'default'

    def query(self):
        return self.host_model_class.objects.using(self.django_dbkey).all()


class BatchImporter(Importer):
    """
    Base class for importers which split their jobs up into batches
    """

    def import_data(self, host_data=None, now=None, **kwargs):
        """
        Import some data!  We must override the default logic here, in order to
        chop up the job into proper batches.  This is because image data is
        relatively large, and fetching all at once is not performant.
        """
        if host_data is not None:
            raise ValueError("User-provided host data is not supported")

        self.now = now or make_utc(tzinfo=True)
        if kwargs:
            self._setup(**kwargs)
        self.setup()
        created = []
        updated = []
        deleted = []

        self.host_index = 0
        host_data = self.normalize_host_data()
        while host_data:
            if self.caches_local_data:
                self.cached_local_data = self.cache_local_data(host_data)

            if self.collect_changes_for_processing:
                created, updated = self._import_create_update(host_data, created=created, updated=updated)
            else:
                self._import_create_update(host_data)

            if self.max_create and len(created) >= self.max_create:
                log.warning("max of {} *created* records has been reached; stopping now".format(self.max_create))
                break
            if self.max_update and len(updated) >= self.max_update:
                log.warning("max of {} *updated* records has been reached; stopping now".format(self.max_update))
                break
            if self.max_total and (len(created) + len(updated)) >= self.max_total:
                log.warning("max of {} *total changes* has been reached; stopping now".format(self.max_total))
                break

            self.host_index += self.batch_size
            host_data = self.normalize_host_data()

        self.teardown()
        # note that these may not be accurate, if we didn't collect changes above
        return created, updated, deleted

    def cache_local_data(self, host_data=None):
        if host_data is None:
            raise ValueError("Must provide host data to this method")
        if len(self.key) != 1:
            raise RuntimeError("Compound key {} not supported for batch importer: {}".format(self.key, self))
        key = self.key[0]
        keys = [data[key] for data in host_data]
        if keys:
            query = self.session.query(self.model_class)\
                                .filter(getattr(self.model_class, key).in_(keys))
            return self.cache_model(self.model_class, key=self.get_cache_key,
                                    query=query, query_options=self.cache_query_options(),
                                    normalizer=self.normalize_cache_object,
                                    message=self.cache_local_message())
        return {} # empty cache


class BulkImporter(Importer):
    """
    Base class for bulk data importers.
    """

    def import_data(self, host_data=None, now=None, **kwargs):
        self.now = now or make_utc(tzinfo=True)
        if kwargs:
            self._setup(**kwargs)
        self.setup()
        if host_data is None:
            host_data = self.normalize_host_data()
        created = self._import_create(host_data)
        self.teardown()
        return created

    def _import_create(self, data):
        count = len(data)
        if not count:
            return 0
        created = count

        prog = None
        if self.progress:
            prog = self.progress("Importing {} data".format(self.model_name), count)

        for i, host_data in enumerate(data, 1):

            key = self.get_key(host_data)
            self.create_object(key, host_data)
            if self.max_create and i >= self.max_create:
                log.warning("max of {} *created* records has been reached; stopping now".format(self.max_create))
                created = i
                break

            # flush changes every so often
            if not self.batch_size or i % self.batch_size == 0:
                self.flush_create_update()

            if prog:
                prog.update(i)
        if prog:
            prog.destroy()

        self.flush_create_update_final()
        return created



# -*- coding: utf-8 -*-
"""add tax

Revision ID: ee253b66b7b
Revises: 2d046576154c
Create Date: 2015-02-25 19:59:37.123167

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'ee253b66b7b'
down_revision = u'2d046576154c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # tax
    op.create_table('tax',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('code', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('rate', sa.Numeric(precision=7, scale=5), nullable=False),
                    sa.PrimaryKeyConstraint('uuid'),
                    sa.UniqueConstraint('code', name=u'tax_uq_code')
                    )
    op.create_table('tax_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('code', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('description', sa.String(length=255), autoincrement=False, nullable=True),
                    sa.Column('rate', sa.Numeric(precision=7, scale=5), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
                    )
    op.create_index(op.f('ix_tax_version_end_transaction_id'), 'tax_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_tax_version_operation_type'), 'tax_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_tax_version_transaction_id'), 'tax_version', ['transaction_id'], unique=False)

    # product.tax
    op.add_column(u'product', sa.Column('tax_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key(u'product_fk_tax', 'product', 'tax', ['tax_uuid'], ['uuid'])
    op.add_column(u'product_version', sa.Column('tax_uuid', sa.String(length=32), autoincrement=False, nullable=True))


def downgrade():

    # product.tax
    op.drop_column(u'product_version', 'tax_uuid')
    op.drop_constraint(u'product_fk_tax', 'product', type_='foreignkey')
    op.drop_column(u'product', 'tax_uuid')

    # tax
    op.drop_index(op.f('ix_tax_version_transaction_id'), table_name='tax_version')
    op.drop_index(op.f('ix_tax_version_operation_type'), table_name='tax_version')
    op.drop_index(op.f('ix_tax_version_end_transaction_id'), table_name='tax_version')
    op.drop_table('tax_version')
    op.drop_table('tax')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands
"""

from __future__ import unicode_literals, absolute_import

import os
import sys
import time
import datetime
import socket
import logging

from rattail.commands import Subcommand
from rattail.time import make_utc, localtime


log = logging.getLogger(__name__)


class DataSync(Subcommand):
    """
    Manage the data sync daemon
    """
    name = 'datasync'
    description = __doc__.strip()

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')

        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        check = subparsers.add_parser('check', help="(DEPRECATED) Check for stale (lingering) changes in queue")
        check.set_defaults(subcommand='check')

        check_queue = subparsers.add_parser('check-queue', help="Check for stale (lingering) changes in queue")
        check_queue.set_defaults(subcommand='check-queue')

        check_watchers = subparsers.add_parser('check-watchers', help="Check for dead watcher threads")
        check_watchers.set_defaults(subcommand='check-watchers')

        collectd = subparsers.add_parser('collectd', help="Output statistics for collectd")
        collectd.set_defaults(subcommand='collectd')

        wait = subparsers.add_parser('wait', help="Wait for changes to be processed")
        wait.set_defaults(subcommand='wait')

        remove_settings = subparsers.add_parser(
            'remove-settings',
            help="Remove all datasync settings from the DB")
        remove_settings.set_defaults(subcommand='remove-settings')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/datasync.pid',
                            help="Path to PID file.")
        parser.add_argument('--daemonize', action='store_true', default=True, # TODO: should default to False
                            help="Daemonize when starting.")
        parser.add_argument('--no-daemonize',
                            '-D', '--do-not-daemonize', # TODO: (re)move these?
                            action='store_false', dest='daemonize',
                            help="Do not daemonize when starting.")
        parser.add_argument('-T', '--timeout', metavar='MINUTES', type=int, default=0,
                            help="Optional timeout (in minutes) for use with the 'wait' or 'check' commands.  "
                            "If specified for 'wait', the waiting still stop after the given number of minutes "
                            "and exit with a nonzero code to indicate failure.  If specified for 'check', the "
                            "command will perform some health check based on the given timeout, and exit with "
                            "nonzero code if the check fails.")

        parser.add_argument('--collectd-plugin', default='datasync',
                            help="The `plugin-instance` part of the 'Identifier' "
                            "to be passed to collectd via PUTVAL entries.")

    def run(self, args):
        from rattail.datasync.daemon import DataSyncDaemon

        if args.subcommand == 'wait':
            self.wait(args)

        elif args.subcommand == 'check':
            self.check_queue(args)

        elif args.subcommand == 'check-queue':
            self.check_queue(args)

        elif args.subcommand == 'check-watchers':
            self.check_watchers(args)

        elif args.subcommand == 'collectd':
            self.collectd(args)

        elif args.subcommand == 'remove-settings':
            self.remove_settings(args)

        else: # manage the daemon
            daemon = DataSyncDaemon(args.pidfile, config=self.config)
            if args.subcommand == 'stop':
                daemon.stop()
            else: # start
                try:
                    daemon.start(daemonize=args.daemonize)
                except KeyboardInterrupt:
                    if not args.daemonize:
                        self.stderr.write("Interrupted.\n")
                    else:
                        raise

    def wait(self, args):
        model = self.model
        session = self.make_session()
        started = make_utc()
        log.debug("will wait for current change queue to clear")
        last_logged = started

        changes = session.query(model.DataSyncChange)
        count = changes.count()
        log.debug("there are %d changes in the queue", count)
        while count:
            try:
                now = make_utc()

                if args.timeout and (now - started).seconds >= (args.timeout * 60):
                    log.warning("datasync wait timed out after %d minutes, with %d changes in queue",
                                args.timeout, count)
                    sys.exit(1)

                if (now - last_logged).seconds >= 60:
                    log.debug("still waiting, %d changes in the datasync queue", count)
                    last_logged = now

                time.sleep(1)
                count = changes.count()

            except KeyboardInterrupt:
                self.stderr.write("Waiting cancelled by user\n")
                session.close()
                sys.exit(1)

        session.close()
        log.debug("all datasync changes have been processed")

    def check_queue(self, args):
        """
        Perform general queue / health check for datasync.
        """
        model = self.model
        session = self.make_session()

        # looking for changes which have been around for "timeout" minutes
        timeout = args.timeout or 90
        cutoff = make_utc() - datetime.timedelta(seconds=60 * timeout)
        changes = session.query(model.DataSyncChange)\
                         .filter(model.DataSyncChange.obtained < cutoff)\
                         .count()
        session.close()

        # if we found stale changes, then "fail" - otherwise we'll "succeed"
        if changes:
            # TODO: should log set of unique payload types, for troubleshooting
            log.debug("found %s changes, in queue for %s minutes", changes, timeout)
            self.stderr.write("Found {} changes, in queue for {} minutes\n".format(changes, timeout))
            sys.exit(1)

        log.info("found no changes in queue for %s minutes", timeout)

    def check_watchers(self, args):
        """
        Perform general health check for datasync watcher threads.
        """
        from rattail.datasync.config import load_profiles
        from rattail.datasync.util import get_lastrun

        profiles = load_profiles(self.config)
        session = self.make_session()

        # cutoff is "timeout" minutes before "now"
        timeout = args.timeout or 15
        cutoff = make_utc() - datetime.timedelta(seconds=60 * timeout)

        dead = []
        for key in profiles:

            # looking for watcher "last run" time older than "timeout" minutes
            lastrun = get_lastrun(self.config, key, tzinfo=False, session=session)
            if lastrun and lastrun < cutoff:
                dead.append(key)

        session.close()

        # if we found dead watchers, then "fail" - otherwise we'll "succeed"
        if dead:
            self.stderr.write("Found {} watcher threads dead for {} minutes: {}\n".format(len(dead), timeout, ', '.join(dead)))
            sys.exit(1)

        log.info("found no watcher threads dead for %s minutes", timeout)

    def collectd(self, args):
        """
        Output statistics for collectd
        """
        from rattail.datasync.config import load_profiles
        from rattail.datasync.util import get_lastrun

        plugin = args.collectd_plugin
        profiles = load_profiles(self.config)
        session = self.app.make_session()
        model = self.model

        # current time (sanity check)
        now = localtime(self.config)
        self.collectd_putval(plugin, 'gauge-current_time',
                             now.timestamp(), now=now)

        # iterate watcher profiles
        for key in profiles:

            # last run for watcher
            lastrun = get_lastrun(self.config, key, local=True, tzinfo=True,
                                  session=session)
            self.collectd_putval(plugin, 'gauge-{}_watcher_lastrun'.format(key),
                                 lastrun.timestamp(), now=now)

            # seconds since last run
            self.collectd_putval(plugin, 'gauge-{}_watcher_seconds_since_lastrun'.format(key),
                                 now.timestamp() - lastrun.timestamp(), now=now)

        # now inspect the change queue..

        # total changes in queue
        total = session.query(model.DataSyncChange).count()
        self.collectd_putval(plugin, 'gauge-queue_changes_total',
                             total, now=now)

        # stale changes in queue
        timeout = self.config.getint('rattail.datasync',
                                     'collectd.changes_stale_timeout',
                                     default=15) # minutes
        cutoff = make_utc() - datetime.timedelta(seconds=60 * timeout)
        stale = session.query(model.DataSyncChange)\
                       .filter(model.DataSyncChange.obtained < cutoff)\
                       .count()
        self.collectd_putval(plugin, 'gauge-queue_changes_stale',
                             stale, now=now)
        self.collectd_putval(plugin, 'gauge-queue_changes_stale_timeout',
                             timeout, now=now)

        session.close()

    def collectd_putval(self, plugin, data_type, value,
                        hostname=None, interval=None, now=None):
        if not hostname:
            hostname = os.environ.get('COLLECTD_HOSTNAME')
            if not hostname:
                hostname = socket.getfqdn()

        if not interval:
            interval = os.environ.get('COLLECTD_INTERVAL')
        if interval:
            interval = ' interval={}'.format(interval)
        else:
            interval = ''

        if now:
            value = '{}:{}'.format(now.timestamp(), value)
        else:
            value = 'N:{}'.format(value)

        msg = 'PUTVAL {}/{}/{}{} {}\n'.format(
            hostname,
            plugin,
            data_type,
            interval,
            value)

        sys.stdout.write(msg)

    def remove_settings(self, args):
        """
        Remove all datasync settings from the DB
        """
        from rattail.datasync.util import purge_datasync_settings

        session = self.app.make_session()
        purge_datasync_settings(self.config, session)
        session.commit()
        session.close()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail "local" data import, for sake of Shopfoo pattern data
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.db import model
from rattail.core import get_uuid


class ShopfooProductImporterMixin(object):
    """
    Product -> ShopfooProduct
    """
    host_model_class = model.Product
    key = 'uuid'

    def setup(self):
        super(ShopfooProductImporterMixin, self).setup()
        self.cache_shopfoo_products()

    def cache_shopfoo_products(self):
        self.shopfoo_products = self.cache_model(self.model_class,
                                                 key='product_uuid')

    def query(self):
        return self.host_session.query(model.Product)\
                                .options(orm.joinedload(model.Product.brand))\
                                .options(orm.joinedload(model.Product.regular_price))

    def identify_shopfoo_product(self, product):
        if hasattr(self, 'shopfoo_products'):
            return self.shopfoo_products.get(product.uuid)

        try:
            return self.session.query(self.model_class)\
                               .filter(self.model_class.product == product)\
                               .one()
        except orm.exc.NoResultFound:
            pass

    def identify_shopfoo_product_uuid(self, product):

        # if we can identify the existing shopfoo product, use its uuid
        shopfoo_product = self.identify_shopfoo_product(product)
        if shopfoo_product:
            return shopfoo_product.uuid

        # otherwise we must make a new one, which means a new record
        return get_uuid()

    def normalize_host_object(self, product):

        # first get the primary data values
        data = self.normalize_base_product_data(product)

        # mark as "unwanted" if we no longer want this product
        unwanted = self.product_is_unwanted(product, data)
        if unwanted:
            data = self.mark_unwanted(product, data)

        data['uuid'] = self.identify_shopfoo_product_uuid(product)
        return data

    def normalize_base_product_data(self, product):
        """
        This method should return a dictionary with *all* primary data field
        values set, for the given product.  Note that a return value is always
        expected from this method; we have separate logic for dealing with
        "unwanted" products, so you should *not* return ``None`` here.
        """
        return {
            'product_uuid': product.uuid,
        }

    def product_is_unwanted(self, product, data):
        """
        Must return a boolean indicating if the product is *unwanted*.  If a
        product is deemed unwanted, and it does not yet exist in Shopfoo, it
        will not be added.  However if it does already exist in Shopfoo, we
        will continue to update it.
        """
        if product.deleted:
            return True
        if product.discontinued:
            return True
        if product.not_for_sale:
            return True
        return False

    def mark_unwanted(self, product, data):
        """
        This method should mark the *data* such that it indicates the product
        is unwanted.  It should not actually try to modify the product.
        """
        data['_deleted_'] = True
        return data



# -*- coding: utf-8; -*-
"""add product_cost.discount

Revision ID: 35b451f8cc27
Revises: adcb91f473c3
Create Date: 2018-11-01 16:45:52.932517

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '35b451f8cc27'
down_revision = 'adcb91f473c3'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product_cost
    op.add_column('product_cost', sa.Column('discount_amount', sa.Numeric(precision=9, scale=5), nullable=True))
    op.add_column('product_cost', sa.Column('discount_cost', sa.Numeric(precision=9, scale=5), nullable=True))
    op.add_column('product_cost', sa.Column('discount_ends', sa.DateTime(), nullable=True))
    op.add_column('product_cost', sa.Column('discount_percent', sa.Numeric(precision=7, scale=4), nullable=True))
    op.add_column('product_cost', sa.Column('discount_starts', sa.DateTime(), nullable=True))
    op.add_column('product_cost_version', sa.Column('discount_amount', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True))
    op.add_column('product_cost_version', sa.Column('discount_cost', sa.Numeric(precision=9, scale=5), autoincrement=False, nullable=True))
    op.add_column('product_cost_version', sa.Column('discount_ends', sa.DateTime(), autoincrement=False, nullable=True))
    op.add_column('product_cost_version', sa.Column('discount_percent', sa.Numeric(precision=7, scale=4), autoincrement=False, nullable=True))
    op.add_column('product_cost_version', sa.Column('discount_starts', sa.DateTime(), autoincrement=False, nullable=True))


def downgrade():

    # product_cost
    op.drop_column('product_cost_version', 'discount_starts')
    op.drop_column('product_cost_version', 'discount_percent')
    op.drop_column('product_cost_version', 'discount_ends')
    op.drop_column('product_cost_version', 'discount_cost')
    op.drop_column('product_cost_version', 'discount_amount')
    op.drop_column('product_cost', 'discount_starts')
    op.drop_column('product_cost', 'discount_percent')
    op.drop_column('product_cost', 'discount_ends')
    op.drop_column('product_cost', 'discount_cost')
    op.drop_column('product_cost', 'discount_amount')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail "local" data import
"""

from __future__ import unicode_literals, absolute_import

from rattail import importing


class FromRattailLocalToRattail(importing.FromRattailHandler, importing.ToSQLAlchemyHandler):
    """
    Common base class for import handlers which read data from the Rattail DB
    for the sake of updating misc. other tables in that same DB.
    """
    host_title = "Rattail (default)"
    local_title = "Rattail (default)"

    def begin_local_transaction(self):
        self.session = self.host_session

        # ToRattailHandler would do this for us, but alas..we copy/pasted this
        if hasattr(self, 'runas_username') and self.runas_username:
            self.session.set_continuum_user(self.runas_username)

    def rollback_transaction(self):
        self.rollback_host_transaction()

    def commit_transaction(self):
        self.commit_host_transaction()


class FromRattailLocal(importing.FromSQLAlchemy):
    """
    Common base class for the "host" side of importers which read data from the
    Rattail DB for the sake of updating misc. other tables in that same DB.
    """



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Windows CE App Interface
"""

from __future__ import unicode_literals, absolute_import

import os
import os.path

from rattail import files
from rattail.csvutil import DictWriter
from rattail.core import Object
from rattail.util import progress_loop


def collect_batch(path, queue, device='Default'):
    """
    This function is meant to be invoked by the file monitor.  Its purpose is
    to gather handheld scan batch files from arbitrary device-specific folders,
    and move them to yet another ("upstream") folder so that they may be part
    of a single processing queue.  Rather than moving the file as-is, the
    following processing is done on the file:

    Since the Rattail CE app saves the batch contents using a somewhat "binary"
    format, the raw file is parsed (with :func:`parse_batch_file()`) and a
    proper CSV file is written with the "translated" contents.  This file is
    then copied to the collection folder.

    The filename used for the final destination path is also manipulated.
    Instead of using the original filename, one is generated which contains the
    original filename, plus the "device name" and the timestamp of the original
    file.  The ``device`` parameter is used to determine the device name.
    """

    ctime = files.creation_time(path)
    final_path = os.path.join(queue, '%s,%s,%s.csv' % (
            device,
            ctime.strftime('%Y-%m-%d %H-%M-%S'),
            os.path.basename(path)))

    csv_path = files.temp_path(suffix='.csv')
    csv_file = open(csv_path, 'wb')
    writer = DictWriter(csv_file, ['upc', 'cases', 'units'])
    writer.writeheader()
    for upc, cases, units in parse_batch_file(path):
        writer.writerow({
                'upc': upc,
                'cases': cases,
                'units': units,
                })
    csv_file.close()

    # TODO: Ugh, this locking copy situation is getting out of hand..FIXME.
    files.locking_copy_old(csv_path, final_path)


def parse_batch_file(path, progress=None):
    """
    Generator which parses a scan batch file originating from the Rattail CE
    application.

    When a valid record is encountered, it is returned as a tuple of the form
    ``(scancode, cases, units)``.
    """
    current = Object(scancode=None,
                     cases=None,
                     units=None)
    results = []

    def parse(line, i):
        line = line.decode('ascii')
        line = line.replace('\x00', '')
        line = line.rstrip('\n')
        
        if current.scancode is None:
            current.scancode = line

        elif current.cases is None:
            line = line.strip()
            if line and line.isdigit():
                current.cases = int(line)
            else:
                current.cases = 0

        elif current.units is None:
            line = line.strip()
            if line and line.isdigit():
                current.units = int(line)
            else:
                current.units = 0

            results.append((current.scancode, current.cases, current.units))
            current.scancode = None
            current.cases = None
            current.units = None

    with open(path, 'rb') as f:
        lines = list(f)

    progress_loop(parse, lines, progress,
                  message="Parsing scan batch file")
    return results



# -*- coding: utf-8 -*-
"""add custorder_item.status_text

Revision ID: 8856f697902d
Revises: 8b78ef45a36c
Create Date: 2021-10-18 11:51:34.326750

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '8856f697902d'
down_revision = '8b78ef45a36c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # custorder_item
    op.add_column('custorder_item', sa.Column('price_needs_confirmation', sa.Boolean(), nullable=True))
    op.add_column('custorder_item', sa.Column('status_text', sa.String(length=255), nullable=True))

    # batch_custorder_row
    op.add_column('batch_custorder_row', sa.Column('price_needs_confirmation', sa.Boolean(), nullable=True))


def downgrade():

    # batch_custorder_row
    op.drop_column('batch_custorder_row', 'price_needs_confirmation')

    # custorder_item
    op.drop_column('custorder_item', 'status_text')
    op.drop_column('custorder_item', 'price_needs_confirmation')



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import logging
from unittest import TestCase
from six import StringIO

import six
from mock import patch

from rattail import logging as rattail_logging


class TestLogging(TestCase):

    @patch('rattail.logging.os')
    @patch('rattail.logging.sys')
    @patch('rattail.logging.socket')
    @patch('rattail.logging.getpass')
    def test_adapter_adds_all_context(self, getpass, socket, sys, os):
        socket.getfqdn.return_value = 'testing.rattailproject.org'
        socket.gethostbyname.return_value = '127.0.0.1'
        sys.argv = ['just', 'testing']
        os.getuid.return_value = 420
        getpass.getuser.return_value = 'joeschmoe'
        formatter = logging.Formatter(u"%(hostname)s %(hostip)s %(argv)s %(username)s %(uid)s %(levelname)s %(message)s")
        string = StringIO()
        handler = logging.StreamHandler(string)
        handler.setFormatter(formatter)
        log = logging.getLogger('fake_for_testing')
        log.addHandler(handler)
        log.propagate = False
        log = rattail_logging.RattailAdapter(log)
        self.assertEqual(string.getvalue(), "")
        log.debug("some random thing")
        if six.PY3:
            self.assertEqual(string.getvalue(), "testing.rattailproject.org 127.0.0.1 ['just', 'testing'] joeschmoe 420 DEBUG some random thing\n")
        else:
            self.assertEqual(string.getvalue(), "testing.rattailproject.org 127.0.0.1 [u'just', u'testing'] joeschmoe 420 DEBUG some random thing\n")
        handler.close()
        string.close()



# -*- coding: utf-8; -*-
"""add custorder.contact_name

Revision ID: c10adeff4117
Revises: 5a256a77e6d0
Create Date: 2021-10-07 11:05:10.561894

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'c10adeff4117'
down_revision = u'5a256a77e6d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_custorder
    op.add_column('batch_custorder', sa.Column('contact_name', sa.String(length=100), nullable=True))

    # custorder
    op.add_column('custorder', sa.Column('contact_name', sa.String(length=100), nullable=True))


def downgrade():
    
    # custorder
    op.drop_column('custorder', 'contact_name')
    
    # batch_custorder
    op.drop_column('batch_custorder', 'contact_name')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete handlers - base class
"""

from __future__ import unicode_literals, absolute_import

import re

import sqlalchemy as sa
from sqlalchemy import orm


class Autocompleter(object):
    """
    Base class and partial default implementation for autocomplete
    handlers.  It is expected that all autocomplete handlers will
    ultimately inherit from this base class, therefore it defines the
    implementation "interface" loosely speaking.  Custom autocomplete
    handlers are welcome to supplement or override this as needed, and
    in fact must do so for certain aspects.

    .. attribute:: autocompleter_key

       The key indicates what "type" of autocompleter this is.  It
       should be a string, e.g. ``'products'``.  It will generally
       correspond to the route names used in Tailbone, though not
       always.  

    .. attribute:: max_results

       If set, should return no more than this many results.  The base
       default *is* set, to 100.  Set to ``None`` to disable limiting
       the number of results.

       The reason to limit results is to avoid situations where the
       query returns many thousands of records, so that's slow anyway,
       but then the browser may well freeze up trying to process it.

       Note that your query probably should be *sorted* somehow, so
       that the e.g. "first 100" results are more relevant.

       Any subclass is free to override this, but you an also set it
       directly on an instance, e.g.::

          autocompleter = app.get_autocompleter('products')
          autocompleter.max_results = 250
          results = autocompleter.autocomplete(session, "apple cider vinegar")
    """
    autocompleter_key = None
    max_results = 100

    def __init__(self, config):
        if not self.autocompleter_key:
            raise NotImplementedError("You must define `autocompleter_key` "
                                      "attribute for handler class: {}".format(
                                          self.__class__))
        self.config = config
        self.app = self.config.get_app()
        self.enum = config.get_enum()
        self.model = config.get_model()

    def get_model_class(self):
        return self.model_class

    @property
    def autocomplete_fieldname(self):
        raise NotImplementedError("You must define `autocomplete_fieldname` "
                                  "attribute for handler class: {}".format(
                                      self.__class__))

    def autocomplete(self, session, term, **kwargs):
        """
        The main reason this class exists.  This method accepts a
        ``term`` (string) argument and will return a sequence of
        matching results.
        """
        term = self.prepare_autocomplete_term(term)
        if not term:
            return []

        data = self.get_autocomplete_data(session, term)
        return self.get_autocomplete_results(data)

    def prepare_autocomplete_term(self, term, **kwargs):
        """
        If necessary, massage the incoming search term for use with
        the autocomplete query.
        """
        return term

    def get_autocomplete_data(self, session, term, **kwargs):
        """
        Collect data for matching results, based on the given search
        term.  This method basically does 2 things:

        First it calls :meth:`make_autocomplete_query()` to get the
        final query, then it invokes the query.

        When invoking the query it will "usually" limit the number of
        results, based on :attr:`max_results`.
        """
        query = self.make_autocomplete_query(session, term)

        # maybe limit number of results for efficiency's sake
        if self.max_results:
            return query[:self.max_results]

        # otherwise just return all of them
        return query.all()

    def make_autocomplete_query(self, session, term, **kwargs):
        """
        Build the complete query from which to obtain search results.
        """
        # we are querying one table (and column) by default
        query = self.make_base_query(session)

        # restrict according to business logic etc. if applicable
        query = self.restrict_autocomplete_query(session, query)

        # filter according to search term(s)
        query = self.filter_autocomplete_query(session, query, term)

        # sort results by something meaningful
        query = self.sort_autocomplete_query(session, query)
        return query

    def make_base_query(self, session):
        """
        Create and return the base ("unfiltered") query from which
        search results will ultimately be obtained.
        """
        model_class = self.get_model_class()
        query = session.query(model_class)
        return query

    def restrict_autocomplete_query(self, session, query, **kwargs):
        """
        Optionally restrict ("pre-filter") the query according to any
        applicable business logic.
        """
        return query

    def filter_autocomplete_query(self, session, query, term):
        """
        Apply the actual "search" filtering and return the query.
        """
        model_class = self.get_model_class()
        column = getattr(model_class, self.autocomplete_fieldname)
        criteria = [column.ilike('%{}%'.format(word))
                    for word in term.split()]
        query = query.filter(sa.and_(*criteria))
        return query

    def sort_autocomplete_query(self, session, query):
        model_class = self.get_model_class()
        column = getattr(model_class, self.autocomplete_fieldname)
        query = query.order_by(column)
        return query

    def get_autocomplete_results(self, data):
        """
        Format the data into a final results set for return to the
        caller.
        """
        return [{'label': self.autocomplete_display(rec),
                 'value': self.autocomplete_value(rec)}
                for rec in data]

    def autocomplete_display(self, obj):
        return getattr(obj, self.autocomplete_fieldname)

    def autocomplete_value(self, obj):
        return obj.uuid


class PhoneMagicMixin(object):
    """
    Mixin for adding "phone number magic" to an otherwise sort of
    normal autocompleter.

    The "magic" is that this will try to search for *either* phone
    number *or* contact name.  If the search term includes at least 4
    digits then it is considered to be a phone number search;
    otherwise it will be considered a name search.
    """
    nondigits_pattern = re.compile(r'\D')

    def make_autocomplete_query(self, session, term, **kwargs):
        """
        This is where the magic happens.  We override this to check if
        the search term resembles a phone number and if so, do a phone
        number search; otherwise a name search.
        """
        column = getattr(self.model_class, self.autocomplete_fieldname)

        # define the base query
        query = session.query(self.model_class)\
                       .options(orm.joinedload(self.model_class.phones))

        # does search term look like a phone number?
        phone_term = self.get_phone_search_term(term)
        if phone_term:

            # yep, so just search for the phone number
            query = query.join(self.phone_model_class,
                               self.phone_model_class.parent_uuid == self.model_class.uuid)
            query = query.filter(sa.func.regexp_replace(self.phone_model_class.number,
                                                        r'\D', '', 'g')\
                                 .like('%{}%'.format(phone_term)))

        else: # term does not look like a phone number

            # so just search by name
            criteria = [column.ilike('%{}%'.format(word))
                        for word in term.split()]
            query = query.filter(sa.and_(*criteria))

        # oh, and sort by something useful
        query = query.order_by(column)

        return query

    def get_phone_search_term(self, term):
        """
        Try to figure out if the given search term represents a whole
        or partial phone number, and if so return just the digits.
        """
        digits = self.nondigits_pattern.sub('', term)
        if digits and len(digits) >= 4:
            return digits

    def get_autocomplete_results(self, data):
        """
        We override the formatting of results, because we want the
        autocomplete results themselves, to appear as "<name> <phone>"
        in the dropdown user sees, but we also want to include the
        ``display`` key which contains just the name.

        The reason for this has to do with how the
        ``tailbone-autocomplete`` comonent works.  The ``display``
        (name only)` will be shown on the button after selection is
        made.
        """
        results = []
        for contact in data:
            phone = contact.first_phone()
            name = getattr(contact, self.autocomplete_fieldname)
            if phone:
                label = "{} {}".format(name, phone.number)
            else:
                label = name
            results.append({'value': contact.uuid,
                            'label': label,
                            'display': name})
        return results



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Shopfoo export
"""

class ProductImporterMixin(object):
    """
    Product data importer mixin
    """

    def setup(self):

        # do normal setup (open output file etc.)
        super(ProductImporterMixin, self).setup()

        # get our "runas" user straight
        assert self.runas_username
        self.host_session.set_continuum_user(self.runas_username)
        self.runas_user = self.host_session.continuum_user

        # make a DB record for the export
        self.init_export(self.host_session)

    def teardown(self):

        # do normal teardown (close output file etc.)
        super(ProductImporterMixin, self).teardown()

        # wrap up the export record, maybe upload file
        self.finalize_export()

    def update_object(self, product, data, local_data=None):
        product = super(ProductImporterMixin, self).update_object(product, data, local_data)

        # keep track of how many records are exported
        self.record_count += 1

        return product



# -*- coding: utf-8; -*-
"""add catalog cost for receiving

Revision ID: fd0975329091
Revises: e838d0b930fc
Create Date: 2019-11-22 19:42:24.126125

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'fd0975329091'
down_revision = 'e838d0b930fc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('catalog_cost_confirmed', sa.Boolean(), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('catalog_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('invoice_cost_confirmed', sa.Boolean(), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('catalog_unit_cost', sa.Numeric(precision=7, scale=3), nullable=True))


def downgrade():

    # purchase_item
    op.drop_column('purchase_item', 'catalog_unit_cost')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'invoice_cost_confirmed')
    op.drop_column('purchase_batch_row', 'catalog_unit_cost')
    op.drop_column('purchase_batch_row', 'catalog_cost_confirmed')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Synchronization for Linux
"""

from __future__ import unicode_literals

from ...daemon import Daemon
from rattail.db.util import get_default_engine
from . import get_sync_engines, synchronize_changes


class SyncDaemon(Daemon):

    def run(self):
        remote_engines = get_sync_engines(self.config)
        if remote_engines:
            local_engine = get_default_engine(self.config)
            synchronize_changes(self.config, local_engine, remote_engines)


def get_daemon(config, pidfile=None):
    """
    Get a :class:`SyncDaemon` instance.
    """

    if pidfile is None:
        pidfile = config.get('rattail.db', 'sync.pid_path',
                             default='/var/run/rattail/dbsync.pid')
    daemon = SyncDaemon(pidfile)
    daemon.config = config
    return daemon


def start_daemon(config, pidfile=None, daemonize=True):
    """
    Start the database synchronization daemon.
    """

    get_daemon(config, pidfile).start(daemonize)


def stop_daemon(config, pidfile=None):
    """
    Stop the database synchronization daemon.
    """

    get_daemon(config, pidfile).stop()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Mail Monitor Daemon
"""

from __future__ import unicode_literals, absolute_import

import re
import time
import imaplib
from six.moves import queue
# import sys
import logging
# from traceback import format_exception

import six

from rattail.daemon import Daemon
from rattail.db import api
from rattail.mailmon.config import load_mailmon_profiles
from rattail.mailmon.util import (get_lastrun,
                                  get_lastrun_setting,
                                  get_lastrun_timefmt)
from rattail.threads import Thread
from rattail.time import make_utc
from rattail.exceptions import StopProcessing
# from rattail.mail import send_email


log = logging.getLogger(__name__)


class MailMonitorDaemon(Daemon):
    """
    Daemon responsible for checking IMAP folders and detecting email
    messages, and then invoking actions upon them.
    """

    def run(self):
        """
        Starts watcher and worker threads according to configuration.
        """
        monitored = load_mailmon_profiles(self.config)
        for key, profile in six.iteritems(monitored):

            # create a msg queue for the profile
            profile.queue = queue.Queue()

            # create a watcher thread for the IMAP folder
            watcher = IMAPWatcher(profile, key)
            name = 'watcher_{}'.format(key)
            log.info("starting IMAP watcher thread: %s", name)
            thread = Thread(target=watcher, name=name)
            thread.daemon = True
            thread.start()

            # create an action thread for the profile
            name = 'actions-{}'.format(key)
            log.debug("starting action thread: %s", name)
            thread = Thread(target=perform_actions, name=name, 
                            args=(self.config, watcher))
            thread.daemon = True
            thread.start()

        # loop indefinitely.  since this is the main thread, the app
        # will terminate when this method ends; all other threads are
        # "subservient" to this one.
        while True:
            time.sleep(.01)


class IMAPWatcher(object):
    """
    Abstraction to make watching an IMAP folder a little more
    organized.  Instances of this class are used as callable targets
    when the daemon starts watcher threads.  They are responsible for
    polling the IMAP folder and processing any messages found there.
    """
    uid_pattern = re.compile(r'^\d+ \(UID (?P<uid>\d+)')

    def __init__(self, profile, key):
        self.profile = profile
        self.config = profile.config
        self.key = key
        self.server = None

        # we need these for writing lastrun times
        self.lastrun_setting = get_lastrun_setting(self.config, self.key)
        self.lastrun_timefmt = get_lastrun_timefmt(self.config)

    def get_uid(self, response):
        match = self.uid_pattern.match(response)
        if match:
            return match.group('uid')

    def __call__(self):
        """
        This is the main loop for the watcher.  It acts as the
        callable target for the thread in which it runs.  It basically
        checks for new messages, queueing any found, then waits for a
        spell, then does it again, forever.
        """
        # the 'lastrun' value is maintained as zone-aware UTC
        lastrun = get_lastrun(self.config, self.key)

        recycled = None
        while True:
            thisrun = make_utc(tzinfo=True)

            if self.server is None:
                self.server = imaplib.IMAP4_SSL(self.profile.imap_server)
                try:
                    result = self.server.login(self.profile.imap_username, self.profile.imap_password)
                except self.server.error:
                    log.exception("failed to login to server!")
                    return

                log.debug("IMAP server login result: %s", result)
                recycled = make_utc()

                result = self.server.select(self.profile.imap_folder)
                log.debug("IMAP server select result: %s", result)

            try:
                self.queue_messages()
            except:
                log.exception("failed to queue messages!")
                if self.profile.stop_on_error:
                    break
            else:
                # successful, so record lastrun time
                lastrun = thisrun
                api.save_setting(None, self.lastrun_setting,
                                 lastrun.strftime(self.lastrun_timefmt))

            # if recycle time limit has been reached, close the IMAP
            # connection; it will be re-established in next loop run
            if self.profile.imap_recycle:
                if (make_utc() - recycled).seconds >= self.profile.imap_recycle:
                    log.debug("recycle time limit reached, disposing of current connection")
                    self.server.close()
                    self.server.logout()
                    self.server = None

            # wait a tick before doing it all again
            time.sleep(self.profile.imap_delay)

        self.server.close()
        self.server.logout()

    def queue_messages(self):
        """
        Check for new messages in the folder, and queue any found, for
        action processing thread.
        """
        # maybe look for "all" or maybe just "unread"
        if self.profile.imap_unread_only:
            criterion = '(UNSEEN)'
        else:
            criterion = 'ALL'

        # log.debug("invoking IMAP4.search()")
        code, items = self.server.uid('search', None, criterion)
        if code != 'OK':
            raise RuntimeError("IMAP4.search() returned bad code: {}".format(code))

        # config may dictacte a "max batch size" in which case we will
        # only queue so many messages at a time
        uids = items[0].split()
        if self.profile.max_batch_size:
            if len(uids) > self.profile.max_batch_size:
                uids = uids[:self.profile.max_batch_size]

        # add message uids to the queue
        for uid in uids:
            self.profile.queue.put(uid)


def perform_actions(config, watcher):
    """
    Target for action threads.  Provides the main loop which checks
    the queue for new messages and invokes actions for each, as they
    appear.
    """
    profile = watcher.profile
    stop = False
    while not stop:

        # suspend execution briefly, to avoid consuming so much CPU...
        time.sleep(0.01)

        try:
            msguid = profile.queue.get_nowait()
        except queue.Empty:
            pass
        except StopProcessing:
            stop = True
        else:
            log.debug("queue contained a msguid: %s", msguid)
            for action in profile.actions:
                try:
                    invoke_action(config, watcher, action, msguid)

                except:
                    # stop processing messages altogether for this
                    # profile if it is so configured
                    if profile.stop_on_error:
                        log.warning("an error was encountered, and config "
                                    "dictates that no more actions should be "
                                    "processed for profile: %s", profile.key)
                        stop = True

                    # either way no more actions should be invoked for
                    # this particular message
                    break


def invoke_action(config, watcher, action, msguid):
    """
    Invoke a single action on a mail message, retrying as necessary.
    """
    attempts = 0
    errtype = None
    while True:
        attempts += 1
        log.debug("invoking action '%s' (attempt #%s of %s) on file: %s",
                  action.spec, attempts, action.retry_attempts, msguid)

        try:
            action.action(watcher.server, msguid, *action.args, **action.kwargs)

        except:

            # if we've reached our final attempt, stop retrying
            if attempts >= action.retry_attempts:
                # log.debug("attempt #%s failed for action '%s' (giving up) on "
                #           "msguid: %s", attempts, action.spec, msguid,
                #           exc_info=True)
                log.exception("attempt #%s failed for action '%s' (giving up) on "
                          "msguid: %s", attempts, action.spec, msguid)
                # TODO: add email support
                # exc_type, exc, traceback = sys.exc_info()
                # send_email(config, 'mailmon_action_error', {
                #     # 'hostname': socket.gethostname(),
                #     # 'path': path,
                #     'msguid': msguid,
                #     'action': action,
                #     'attempts': attempts,
                #     'error': exc,
                #     'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
                # })
                raise

            # if this exception is not the first, and is of a
            # different type than seen previously, do *not* continue
            # to retry
            if errtype is not None and not isinstance(error, errtype):
                log.exception("new exception differs from previous one(s), "
                              "giving up on action '%s' for msguid: %s",
                              action.spec, msguid)
                raise

            # record the type of exception seen, and pause for next retry
            log.warning("attempt #%s failed for action '%s' on msguid: %s",
                        attempts, action.spec, msguid, exc_info=True)
            errtype = type(error)
            log.debug("pausing for %s seconds before making attempt #%s of %s",
                      action.retry_delay, attempts + 1, action.retry_attempts)
            if action.retry_delay:
                time.sleep(action.retry_delay)

        else:
            # no error, invocation successful
            log.debug("attempt #%s succeeded for action '%s' on msguid: %s",
                      attempts, action.spec, msguid)
            break



# -*- coding: utf-8; -*-
"""add merge_request_people

Revision ID: 51f7773bb07a
Revises: 9072c76a3afc
Create Date: 2021-08-19 10:24:09.801411

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '51f7773bb07a'
down_revision = u'9072c76a3afc'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # merge_request_people
    op.create_table('merge_request_people',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('removing_uuid', sa.String(length=32), nullable=False),
                    sa.Column('keeping_uuid', sa.String(length=32), nullable=False),
                    sa.Column('requested', sa.DateTime(), nullable=False),
                    sa.Column('requested_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('merged', sa.DateTime(), nullable=True),
                    sa.Column('merged_by_uuid', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['merged_by_uuid'], [u'user.uuid'], name=u'merge_request_people_fk_merged_by'),
                    sa.ForeignKeyConstraint(['requested_by_uuid'], [u'user.uuid'], name=u'merge_request_people_fk_requested_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # merge_request_people
    op.drop_table('merge_request_people')



# -*- coding: utf-8; -*-
# -*- coding: utf-8; -*-
"""
Handler for ${model_title} batches
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model
from rattail.batch import BatchHandler


class ${model_name}Handler(BatchHandler):
    """
    Handler for ${model_title} batches.
    """
    batch_model_class = model.${model_name}

    def refresh_row(self, row):
        """
        Inspect a single row from the batch, and set its attributes based on
        various lookups and business rules, as needed.
        """

    def execute(self, batch, **kwargs):
        """
        Execute the batch, whatever that means in your context.
        """
        return True



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Auth Handler

See also :doc:`rattail-manual:base/handlers/other/auth`.
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy_continuum as continuum

from rattail.app import GenericHandler


class AuthHandler(GenericHandler):
    """
    Base class and default implementation for auth handlers.
    """

    def authenticate_user(self, session, username, password):
        """
        Try to authenticate the given credentials.  If successful,
        this should return a Rattail ``User`` instance; otherwise
        returns ``None``.
        """
        from rattail.db.auth import authenticate_user

        return authenticate_user(session, username, password)

    def generate_unique_username(self, session, **kwargs):
        """
        Generate a unique username using data from ``kwargs`` as hints.
        """
        model = self.model

        original_username = self.generate_username(session, **kwargs)
        username = original_username

        # only if given a session, can we check for unique username
        if session:
            counter = 1
            while True:
                users = session.query(model.User)\
                               .filter(model.User.username == username)\
                               .count()
                if not users:
                    break
                username = "{}{:02d}".format(original_username, counter)
                counter += 1

        return username

    def generate_username(self, session, **kwargs):
        """
        Generate a unique username using data from ``kwargs`` as hints.
        """
        person = kwargs.get('person')
        if person:
            first = (person.first_name or '').strip().lower()
            last = (person.last_name or '').strip().lower()
            return '{}.{}'.format(first, last)
            
        return 'newuser'

    def make_user(self, **kwargs):
        """
        Make and return a new User instance.
        """
        model = self.model
        session = kwargs.pop('session', None)

        if 'username' not in kwargs:
            kwargs['username'] = self.generate_unique_username(session, **kwargs)

        user = model.User(**kwargs)
        return user

    def grant_permission(self, role, permission):
        """
        Grant a permission to the role.
        """
        if permission not in role.permissions:
            role.permissions.append(permission)

    def revoke_permission(self, role, permission):
        """
        Revoke a permission from the role.
        """
        if permission in role.permissions:
            role.permissions.remove(permission)

    def cache_permissions(self, session, principal,
                          include_guest=True,
                          include_authenticated=True):
        """
        Return a set of permission names, which represents all
        permissions effectively granted to the given principal.

        :param session: A SQLAlchemy session instance.

        :param principal: May be either a
           :class:`~rattail.db.model.users.User` or
           :class:`~rattail.db.model.users.Role` instance.  It is also
           expected that this may sometimes be ``None``, in which case
           the "Guest" role will typically be assumed.

        :param include_guest: Whether or not the "Guest" role should
           be included when checking permissions.  If ``False``, then
           Guest's permissions will *not* be consulted.

        :param include_authenticated: Whether or not the
           "Authenticated" role should be included when checking
           permissions.

        Note that if no ``principal`` is provided, and
        ``include_guest`` is set to ``False``, then no checks will
        actually be done, and the return value will be ``False``.
        """
        from rattail.db.auth import guest_role, authenticated_role

        # we will use any `roles` attribute which may be present.  in practice we
        # would be assuming a User in this case
        if hasattr(principal, 'roles'):

            roles = []
            for role in principal.roles:
                include = False
                if role.node_type:
                    if role.node_type == self.config.node_type():
                        include = True
                else:
                    include = True
                if include:
                    roles.append(role)

            # here our User assumption gets a little more explicit
            if include_authenticated:
                roles.append(authenticated_role(session))

        # otherwise a non-null principal is assumed to be a Role
        elif principal is not None:
            roles = [principal]

        # fallback assumption is "no roles"
        else:
            roles = []

        # maybe include guest roles
        if include_guest:
            roles.append(guest_role(session))

        # build the permissions cache
        cache = set()
        for role in roles:
            cache.update(role.permissions)

        return cache

    def has_permission(self, session, principal, permission,
                       include_guest=True,
                       include_authenticated=True):
        """
        Determine if a principal has been granted a permission.

        Under the hood this really just invokes
        :meth:`cache_permissions()` for the principal, and then checks
        to see if the given permission is included in that set.
        """
        perms = self.cache_permissions(session, principal,
                                       include_guest=include_guest,
                                       include_authenticated=include_authenticated)
        return permission in perms

    def delete_user(self, user, **kwargs):
        """
        Delete the given user account.  Use with caution!  As this
        generally cannot be undone.

        Default behavior here is of course to delete the account, but
        it also must try to "remove" the user association from various
        places, in particular the continuum transactions table.
        Please note that this will leave certain record versions as
        appearing to be "without an author".
        """
        session = self.app.get_session(user)

        # disassociate user from transactions
        self.remove_user_from_continuum_transactions(user)

        # finally, delete the user outright
        session.delete(user)
        return True

    def remove_user_from_continuum_transactions(self, user):
        """
        Remove the given user from all Continuum transactions.
        """
        session = self.app.get_session(user)
        model = self.model

        # remove the user from any continuum transactions
        # nb. we can use "any" model class here, to obtain Transaction
        Transaction = continuum.transaction_class(model.User)
        transactions = session.query(Transaction)\
                              .filter(Transaction.user_id == user.uuid)\
                              .all()
        for txn in transactions:
            txn.user_id = None



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Customer Orders
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import Base, uuid_column
from rattail.db.model import Store, Customer, PendingCustomer, Person, Product, User, Note
from rattail.db.types import GPCType


class CustomerOrderBase(object):
    """
    Base class for customer orders; defines common fields.
    """

    @declared_attr
    def __table_args__(cls):
        return cls.__customer_order_table_args__()

    @classmethod
    def __customer_order_table_args__(cls):
        table_name = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['store_uuid'], ['store.uuid'],
                                    name='{}_fk_store'.format(table_name)),
            sa.ForeignKeyConstraint(['customer_uuid'], ['customer.uuid'],
                                    name='{}_fk_customer'.format(table_name)),
            sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'],
                                    name='{}_fk_person'.format(table_name)),
            sa.ForeignKeyConstraint(['pending_customer_uuid'], ['pending_customer.uuid'],
                                    name='{}_fk_pending_customer'.format(table_name)),
        )

    store_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def store(cls):
        return orm.relationship(
            Store,
            doc="""
            Reference to the store to which the order applies.
            """)

    customer_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def customer(cls):
        return orm.relationship(
            Customer,
            doc="""
            Reference to the customer account for which the order exists.
            """)

    person_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def person(cls):
        return orm.relationship(
            Person,
            doc="""
            Reference to the person to whom the order applies.
            """)

    pending_customer_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def pending_customer(cls):
        return orm.relationship(
            PendingCustomer,
            doc="""
            Reference to the *pending* customer account for the order,
            if applicable.
            """)

    contact_name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Cached display name for the contact (customer).
    """)

    phone_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    Customer contact phone number for sake of this order.
    """)

    email_address = sa.Column(sa.String(length=255), nullable=True, doc="""
    Customer contact email address for sake of this order.
    """)

    total_price = sa.Column(sa.Numeric(precision=10, scale=3), nullable=True, doc="""
    Full price (not including tax etc.) for all items on the order.
    """)


@six.python_2_unicode_compatible
class CustomerOrder(CustomerOrderBase, Base):
    """
    Represents an order placed by the customer.
    """
    __tablename__ = 'custorder'

    @declared_attr
    def __table_args__(cls):
        return cls.__customer_order_table_args__() + (
            sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'],
                                    name='custorder_fk_created_by'),
        )

    uuid = uuid_column()

    id = sa.Column(sa.Integer(), doc="""
    Numeric, auto-increment ID for the order.
    """)

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date and time when the order/batch was first created.
    """)

    created_by_uuid = sa.Column(sa.String(length=32), nullable=True)
    created_by = orm.relationship(
        User,
        doc="""
        Reference to the user who initially created the order/batch.
        """)

    status_code = sa.Column(sa.Integer(), nullable=False)

    items = orm.relationship(
        'CustomerOrderItem',
        back_populates='order',
        collection_class=ordering_list('sequence', count_from=1),
        cascade='all, delete-orphan',
        doc="""
        Sequence of :class:`CustomerOrderItem` instances which belong to the order.
        """)

    def __str__(self):
        if self.id:
            return "#{}".format(self.id)
        return "(pending)"


@six.python_2_unicode_compatible
class CustomerOrderItemBase(object):
    """
    Base class for customer order line items.
    """

    @declared_attr
    def __table_args__(cls):
        return cls.__customer_order_item_table_args__()

    @classmethod
    def __customer_order_item_table_args__(cls):
        table_name = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='{}_fk_product'.format(table_name)),
        )

    product_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def product(cls):
        return orm.relationship(
            Product,
            doc="""
            Reference to the master product record for the line item.
            """)

    product_upc = sa.Column(GPCType(), nullable=True, doc="""
    UPC for the product associated with the row.
    """)

    product_brand = sa.Column(sa.String(length=100), nullable=True, doc="""
    Brand name for the product being ordered.  This should be a cache of the
    relevant :attr:`Brand.name`.
    """)

    product_description = sa.Column(sa.String(length=60), nullable=True, doc="""
    Primary description for the product being ordered.  This should be a cache
    of :attr:`Product.description`.
    """)

    product_size = sa.Column(sa.String(length=30), nullable=True, doc="""
    Size of the product being ordered.  This should be a cache of
    :attr:`Product.size`.
    """)

    product_weighed = sa.Column(sa.String(length=4), nullable=True, doc="""
    Flag indicating whether the product is sold by weight.  This should be a
    cache of :attr:`Product.weighed`.
    """)

    # TODO: probably should get rid of this, i can't think of why it's needed.
    # for now we just make sure it is nullable, since that wasn't the case.
    product_unit_of_measure = sa.Column(sa.String(length=4), nullable=True, doc="""
    Code indicating the unit of measure for the product.  This should be a
    cache of :attr:`Product.unit_of_measure`.
    """)

    department_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of the department to which the product belongs.
    """)

    department_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name of the department to which the product belongs.
    """)

    case_quantity = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Case pack count for the product being ordered.  This should be a cache of
    :attr:`Product.case_size`.
    """)

    # TODO: i now think that cases_ordered and units_ordered should go away.
    # but will wait until that idea has proven itself before removing.  am
    # pretty sure they are obviated by order_quantity and order_uom.

    cases_ordered = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of cases of product which were initially ordered.
    """)

    units_ordered = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Number of units of product which were initially ordered.
    """)

    order_quantity = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Quantity being ordered by the customer.
    """)

    order_uom = sa.Column(sa.String(length=4), nullable=True, doc="""
    Code indicating the unit of measure for the order itself.  Does not
    directly reflect the :attr:`~rattail.db.model.Product.unit_of_measure`.
    """)

    product_unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Unit cost of the product being ordered.  This should be a cache of the
    relevant :attr:`rattail.db.model.ProductCost.unit_cost`.
    """)

    unit_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Unit price for the product being ordered.  This is the price which is
    quoted to the customer and/or charged to the customer, but for a unit only
    and *before* any discounts are applied.  It generally will be a cache of
    the relevant :attr:`ProductPrice.price`.
    """)

    discount_percent = sa.Column(sa.Numeric(precision=5, scale=3), nullable=False, default=0, doc="""
    Discount percentage which will be applied to the product's price as part of
    calculating the :attr:`total_price` for the item.
    """)

    total_price = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
    Full price (not including tax etc.) which the customer is asked to pay for the item.
    """)

    price_needs_confirmation = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that the price for this item should be confirmed
    by someone, before the order advances to the procurement phase.

    Items/rows with this flag set will probably indicate that also via
    their status.

    When the price is eventually confirmed by someone, this flag
    should be cleared and probably the status will update as well.
    """)

    paid_amount = sa.Column(sa.Numeric(precision=8, scale=3), nullable=False, default=0, doc="""
    Amount which the customer has paid toward the :attr:`total_price` of theitem.
    """)

    payment_transaction_number = sa.Column(sa.String(length=8), nullable=True, doc="""
    Transaction number in which payment for the order was taken, if applicable.
    """)

    def __str__(self):
        return str(self.product or "(no product)")


class CustomerOrderItem(CustomerOrderItemBase, Base):
    """
    Represents a particular line item (product) within a customer order.
    """
    __tablename__ = 'custorder_item'

    @declared_attr
    def __table_args__(cls):
        return cls.__customer_order_item_table_args__() + (
            sa.ForeignKeyConstraint(['order_uuid'], ['custorder.uuid'],
                                    name='custorder_item_fk_order'),
        )

    uuid = uuid_column()

    order_uuid = sa.Column(sa.String(length=32), nullable=False)
    order = orm.relationship(CustomerOrder, back_populates='items', doc="""
    Reference to the :class:`CustomerOrder` instance to which the item belongs.
    """)

    sequence = sa.Column(sa.Integer(), nullable=False, doc="""
    Numeric sequence for the item, i.e. its "line number".  These values should
    obviously increment in sequence and be unique within the context of a
    single order.
    """)

    status_code = sa.Column(sa.Integer(), nullable=False)

    status_text = sa.Column(sa.String(length=255), nullable=True, doc="""
    Text which may briefly explain the batch status code, if needed.
    """)

    def add_event(self, type_code, user, **kwargs):
        """
        Convenience method to add an event for the order item.
        """
        self.events.append(CustomerOrderItemEvent(type_code=type_code,
                                                  user=user,
                                                  **kwargs))


class CustomerOrderItemEvent(Base):
    """
    An event in the life of a customer order item
    """
    __tablename__ = 'custorder_item_event'
    __table_args__ = (
        sa.ForeignKeyConstraint(['item_uuid'], ['custorder_item.uuid'], name='custorder_item_event_fk_item'),
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='custorder_item_event_fk_user'),
    )

    uuid = uuid_column()

    item_uuid = sa.Column(sa.String(length=32), nullable=False)

    item = orm.relationship(
        CustomerOrderItem,
        doc="""
        Reference to the :class:`CustomerOrder` instance to which the item belongs.
        """,
        backref=orm.backref(
            'events',
            order_by='CustomerOrderItemEvent.occurred'))

    type_code = sa.Column(sa.Integer, nullable=False)

    occurred = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    Date and time when the event occurred.
    """)

    user_uuid = sa.Column(sa.String(length=32), nullable=False)

    user = orm.relationship(
        User,
        doc="""
        User who was the "actor" for the event.
        """)

    note = sa.Column(sa.Text(), nullable=True, doc="""
    Optional note recorded for the event.
    """)


class CustomerOrderItemNote(Note):
    """
    Represents a note attached to an order item.
    """
    __mapper_args__ = {'polymorphic_identity': 'CustomerOrderItem'}

CustomerOrderItem.notes = orm.relationship(
    CustomerOrderItemNote,
    primaryjoin=CustomerOrderItemNote.parent_uuid == CustomerOrderItem.uuid,
    foreign_keys=[CustomerOrderItemNote.parent_uuid],
    order_by=CustomerOrderItemNote.created,
    cascade='all, delete-orphan',
    doc="""
    Sequence of notes attached to the order item.
    """,
    backref=orm.backref(
        'person',
        doc="""
        Reference to the order item to which the note is attached.
        """))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
"Generic" Catalog Parser
"""

import datetime
import decimal
import logging

import six

from rattail.vendors.catalogs import CatalogParser
from rattail.excel import ExcelReaderXLSX
from rattail.db import model
from rattail.gpc import GPC
from rattail.time import localtime, make_utc


log = logging.getLogger(__name__)


class GenericCatalogParser(CatalogParser):
    """
    Generic vendor catalog parser, for Excel XLSX files.
    """
    key = 'rattail.contrib.generic'
    display = "Generic Excel (XLSX only)"

    def parse_rows(self, path, progress=None):
        """
        This parser expects a "standard" XLSX file with one header row.
        """
        reader = ExcelReaderXLSX(path)
        xlrows = reader.read_rows(progress=progress)

        for xlrow in xlrows:
            row = model.VendorCatalogBatchRow()

            # upc (required)
            upc = xlrow['UPC']
            if not upc or not six.text_type(upc).strip():
                continue        # skip lines with no UPC value
            upc = str(upc).replace(' ', '').replace('-', '')
            row.upc = GPC(upc, calc_check_digit=False if len(upc) in (12, 13) else 'upc')

            # cost values (required in some combination)
            if 'Case Cost' in xlrow:
                case_cost = xlrow['Case Cost']
                if isinstance(case_cost, int):
                    row.case_cost = decimal.Decimal('{}.00'.format(case_cost))
                else:
                    row.case_cost = self.decimal(case_cost)
            if 'Case Size' in xlrow:
                row.case_size = int(xlrow['Case Size'])
            if 'Unit Cost' in xlrow:
                row.unit_cost = self.decimal(xlrow['Unit Cost'])
            elif row.case_cost and row.case_size:
                row.unit_cost = row.case_cost / row.case_size

            # optional values
            if 'Vendor Code' in xlrow:
                row.vendor_code = six.text_type(xlrow['Vendor Code'])
            if 'Brand' in xlrow:
                row.brand_name = xlrow['Brand']
            if 'Description' in xlrow:
                row.description = xlrow['Description']
            if 'Unit Size' in xlrow:
                row.size = xlrow['Unit Size']

            if 'SRP' in xlrow:
                value = xlrow['SRP']
                try:
                    row.suggested_retail = self.decimal(value)
                except decimal.InvalidOperation:
                    log.warning("cannot parse SRP value: %s", value)

            # discount_starts
            if 'Vendor Discount Start Date' in xlrow:
                row.discount_starts = xlrow['Vendor Discount Start Date']
                if row.discount_starts: # must convert to UTC, at local midnight
                    date = localtime(self.config, row.discount_starts).date()
                    starts = datetime.datetime.combine(date, datetime.time(0))
                    starts = localtime(self.config, starts)
                    row.discount_starts = make_utc(starts)

            # discount_ends
            if 'Vendor Discount End Date' in xlrow:
                row.discount_ends = xlrow['Vendor Discount End Date']
                if row.discount_ends: # must convert to UTC, 1 minute shy of local midnight
                    date = localtime(self.config, row.discount_ends).date()
                    ends = datetime.datetime.combine(date, datetime.time(23, 59))
                    ends = localtime(self.config, ends)
                    row.discount_ends = make_utc(ends)

            # discount_amount
            if 'Vendor Discount Amount' in xlrow:
                row.discount_amount = self.decimal(xlrow['Vendor Discount Amount'])

            # discount_percent
            if 'Vendor Discount Percent' in xlrow:
                row.discount_percent = self.decimal(xlrow['Vendor Discount Percent'])

            yield row
